#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions related with file operations,
such as read and write files, create and delete files,
manipulate file paths, compress files, verify file contents,
etc.

Code documentation
------------------
"""


import os
import sys
import csv
import math
import time
import gzip
import shutil
import pickle
import zipfile
import urllib.request
from itertools import islice
from multiprocessing import TimeoutError
from multiprocessing.pool import ThreadPool

import pandas as pd

try:
    from utils import (constants as ct,
                       iterables_manipulation as im)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 iterables_manipulation as im)


def file_basename(file_path, file_extension=True, delimiter='.'):
    """Extract file basename from a path.

    Parameters
    ----------
    file_path : str
        Path to the file.
    file_extension : bool
        Specify if the basename should include the file
        extension.
    delimiter : str
        Delimiter used to split the basename to exclude
        the file extension.

    Returns
    -------
    basename : str
        File basename extracted from input path.
    """
    basename = os.path.basename(file_path)

    if file_extension is False:
        basename = basename.split(delimiter)[0]

    return basename


def get_locus_id(locus_path):
    """Extract the locus identifier from a path.

    Parameters
    ----------
    locus_path : str
        Path to the locus Fasta file.

    Returns
    -------
    locus_id : str
        Locus identifier without the .fasta extension.
    """
    locus_basename = file_basename(locus_path)
    locus_id = im.match_regex(locus_basename, ct.LOCUS_ID_PATTERN)

    return locus_id


def remove_files(files):
    """Delete a list of files.

    Parameters
    ----------
    files : list
        List with paths to the files to be deleted.
    """
    for f in files:
        if os.path.isfile(f) is True:
            os.remove(f)


def hash_file(file, hash_object, buffer_size=65536):
    """Compute hash based on the contents of a file.

    Parameters
    ----------
    file : str
        Path to a file.
    hash_object : _hashlib.HASH
        Hashlib object to update based on file
        contents.
    buffer_size : int
        Buffer size (amount of data, in KB, to
        read from file).

    Returns
    -------
    hash_str : str
        Hash computed from file contents.
    """
    updated_hash = hash_object

    with open(file, 'rb') as f:
        # read file in chunks
        while True:
            data = f.read(buffer_size)
            if not data:
                break
            updated_hash.update(data)

    hash_str = updated_hash.hexdigest()

    return hash_str


def filter_files(files, suffixes, reverse=False):
    """Filter files names based on a list of suffixes.

    Parameters
    ----------
    files : list
        A list with filenames or file paths.
    suffixes : list
        List with suffixes.
    reverse : bool
        False to select files that contain any of the suffixes.
        True to select files that do not contain any of the
        suffixes.

    Returns
    -------
    filtered : list
        List with files that passed filtering.
    """
    filtered = [file for file in files
                if any([True for suffix in suffixes if suffix in file])]

    if reverse is True:
        filtered = list(set(files)-set(filtered))

    return filtered


def create_directory(directory_path):
    """Create a diretory if it does not exist."""
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)
        return True
    else:
        return False


def join_paths(parent_path, child_paths):
    """Create path by joining a parent directory and a list of child paths."""
    joined_paths = os.path.join(parent_path, *child_paths)

    return joined_paths


def listdir_fullpath(directory_path, substring_filter=False):
    """Get the full path for all files in a directory.

    Parameters
    ----------
    directory_path : str
        Path to a directory.
    substring_filter : str
        Only list files that contain this substring.

    Returns
    -------
    file_list : list
        List containing the full path for every selected
        file in the input directory.
    """
    file_list = [os.path.join(directory_path, f)
                 for f in os.listdir(directory_path)]

    if substring_filter is not False:
        file_list = filter_files(file_list, [substring_filter])

    return file_list


def delete_directory(directory_path, max_retries=5):
    """Delete a directory."""
    for i in range(max_retries):
        # might fail to delete files due to permission issues
        try:
            shutil.rmtree(directory_path)
        # path does not exist
        # FileNotFoundError is a subclass of OSError
        # the latter will catch the former if checked first
        except FileNotFoundError:
            break
        # sleep before retry
        except OSError:
            time.sleep(i)

    # check if directory still exists
    exists = os.path.isdir(directory_path)
    if exists is True:
        print('Could not remove {0}'.format(directory_path))

    return exists


def read_lines(input_file, strip=True, num_lines=None):
    """Read lines in a file.

    Parameters
    ----------
    input_file : str
        Path to the input file.
    strip : bool
        Specify if lines should be stripped of leading
        and trailing white spaces and new line characters.

    Returns
    -------
    lines : list
        List with the lines read from the input file.
    """
    with open(input_file, 'r') as infile:
        if num_lines is None:
            lines = [line for line in infile.readlines()]
        else:
            lines = list(islice(infile, num_lines))

    if strip is True:
        lines = [line.strip() for line in lines]

    return lines


def pickle_dumper(content, output_file):
    """Use the Pickle module to serialize an object.

    Parameters
    ----------
    content : type
        Variable that refers to the object that will
        be serialized and written to the output file.
    output_file : str
        Path to the output file.
    """
    with open(output_file, 'wb') as poutfile:
        pickle.dump(content, poutfile)


def pickle_loader(input_file):
    """Use the Pickle module to de-serialize an object.

    Parameters
    ----------
    input_file : str
        Path to file with byte stream to be de-serialized.

    Returns
    -------
    content : type
        Variable that refers to the de-serialized
        object.
    """
    with open(input_file, 'rb') as pinfile:
        content = pickle.load(pinfile)

    return content


def file_zipper(input_file, zip_file):
    """Zip (compresses) a file.

    Parameters
    ----------
    input_file : str
        Path to the file that will be compressed.
    zip_file : str
        Path to the ZIP file that will be created.

    Returns
    -------
    zip_file : str
        Path to the ZIP file that was created by
        compressing the input file.
    """
    with zipfile.ZipFile(zip_file, 'w', compression=zipfile.ZIP_DEFLATED) as zf:
        zf.write(input_file, os.path.basename(input_file))

    return zip_file


def unzip_file(compressed_file, archive_type='.gz'):
    """Uncompresse a file.

    Parameters
    ----------
    compressed_file : str
        Path to the compressed file.
    archive_type : str
        Archive format.

    Returns
    -------
    uncompressed_file : str
        Path to the uncompressed file.
    """
    lines = []
    with gzip.open(compressed_file, 'rb') as f:
        for line in f:
            lines.append(line.decode())

    # save uncompressed contents
    uncompressed_file = compressed_file.rstrip('.gz')
    write_lines(lines, uncompressed_file, joiner='')

    return uncompressed_file


def download_file(file_url, outfile, max_tries=3):
    """Download a file.

    Parameters
    ----------
    file_url : str
        URL to download file.
    outfile : str
        Path to the downloaded file.
    max_tries : int
        Maximum number of retries if the download
        fails.
    """
    tries = 0
    downloaded = False
    while downloaded is False and tries <= max_tries:
        try:
            res = urllib.request.urlretrieve(file_url, outfile)
            if os.path.isfile(outfile) is True:
                downloaded = True
        except Exception as e:
            print(e)
            time.sleep(1)
        tries += 1

    return downloaded


def concatenate_files(files, output_file, header=None):
    """Concatenate files.

    Parameters
    ----------
    files : list
        List with the paths to the files to concatenate.
    output_file : str
        Path to the output file that will store the concatenation
        of input files.
    header : str or NoneType
        Specify a header that should be written as the first line
        in the output file.

    Returns
    -------
    output_file : str
        Path to the output file that was created with
        the concatenation of input files.
    """
    with open(output_file, 'w') as outfile:
        if header is not None:
            outfile.write(header)
        for file in files:
            with open(file, 'r') as infile:
                shutil.copyfileobj(infile, outfile)

    return output_file


def write_to_file(text, output_file, write_mode, end_char):
    """Write a single string to a file.

    Parameters
    ----------
    text : str
        A single string to write to the output file.
    output_file : str
        Path to the output file.
    write_mode : str
        Specify write mode ('w' creates file if it does not
        exist and truncates and over-writes existing file,
        'a' creates file if it does not exist and appends to
        the end of file if it exists.).
    end_char : str
        Character added to the end of the file.
    """
    with open(output_file, write_mode) as out:
        out.write(text+end_char)


def write_lines(lines, output_file, joiner='\n', write_mode='w'):
    """Write a list of strings to a file.

    Parameters
    ----------
    lines : list
        List with the lines/strings to write to the output
        file.
    output_file : str
        Path to the output file.
    joiner : str
        Character used to join lines.
    write_mode : str
        Specify write mode ('w' creates file if it does not
        exist and truncates and over-writes existing file,
        'a' creates file if it does not exist and appends to
        the end of file if it exists).
    """
    joined_lines = im.join_list(lines, joiner)

    write_to_file(joined_lines, output_file, write_mode, '\n')


def read_tabular(input_file, delimiter='\t'):
    """Read a tabular (TSV) file.

    Parameters
    ----------
    input_file : str
        Path to a tabular file.
    delimiter : str
        Delimiter used to separate file fields.

    Returns
    -------
    lines : list
        A list with a sublist per line in the input file.
        Each sublist has the fields that were separated by
        the defined delimiter.
    """
    with open(input_file, 'r') as infile:
        reader = csv.reader(infile, delimiter=delimiter)
        lines = [line for line in reader]

    return lines


def input_timeout(prompt, timeout=30):
    """Add timeout feature when requesting user input.

    Parameters
    ----------
    prompt : str
        Message to print to stdout to request for user
        input.
    timeout : int
        Maximum number of seconds that the process will
        wait for input.

    Returns
    -------
    String with user input.

    Raises
    ------
    SystemExit
        - If there is no user input before timeout.
    """
    pool = ThreadPool(processes=1)
    answer = pool.apply_async(input, args=[prompt])

    try:
        return answer.get(timeout=timeout)
    except TimeoutError:
        sys.exit('Timed out.')


def is_file_empty(file_path):
    """Check if file is empty by confirming if its size is 0 bytes."""
    # Check if file exist and it is empty
    return os.path.exists(file_path) and os.stat(file_path).st_size == 0


def is_file_empty_2(file_name):
    """Check if file is empty by confirming if its size is 0 bytes."""
    # Check if file exist and it is empty
    return os.path.isfile(file_name) and os.path.getsize(file_name) == 0


def is_file_empty_3(file_name):
    """Check if file is empty by reading first character in it."""
    # open ile in read mode
    with open(file_name, 'r') as read_obj:
        # read first character
        one_char = read_obj.read(1)
        # if not fetched then file is empty
        if not one_char:
            return True
        elif one_char == " ":
            return True
    return False


def create_short(schema_files, schema_dir):
    """Create the 'short' directory for a chewBBACA schema.

    Creates the 'short' directory inside a schema's main
    directory and copies the FASTA files with the representative
    alleles into the 'short' folder.

    Parameters
    ----------
    schema_files : list
        List with paths to FASTA files that contain the loci
        representative alleles.
    schema_dir : str
        Path to the schema's directory.

    Returns
    -------
    True on completion.
    """
    short_path = join_paths(schema_dir, ['short'])
    create_directory(short_path)

    for file in schema_files:
        short_file = join_paths(short_path, [file_basename(file)])
        short_file = short_file.replace('.fasta', '_short.fasta')
        shutil.copy(file, short_file)

    return True


def move_file(source, destination):
    """Move a file to specified destination."""
    shutil.move(source, destination)


def matching_lines(input_file, pattern):
    """Retrieve lines from a file that contain a pattern.

    Parameters
    ----------
    input_file : str
        Path to input file.
    pattern : str
        Pattern to match.

    Returns
    -------
    matched_lines : list
        List with lines that contain the pattern.
    """
    with open(input_file, 'r') as infile:
        matched_lines = [line for line in infile if pattern in line]

    return matched_lines


def transpose_matrix(input_file, output_directory):
    """Transpose a TSV file.

    Parameters
    ----------
    input_file : str
        Path to the input TSV file.
    output_directory : str
        Path to the directory to which intermediate files
        and the complete transposed file will be written.

    Returns
    -------
    transposed_file : str
        Path to the file with the transposed matrix.
        This file is created by concatenating all
        intermediate files.
    """
    intermediate_files = []
    with open(input_file, 'r') as infile:
        # get column identifiers
        columns = [c.strip() for c in (infile.__next__()).split('\t')]
        # divide into smaller sets to avoid loading complete file
        total_column_sets = math.ceil(len(columns)/500)
        column_sets = im.divide_list_into_n_chunks(columns, total_column_sets)
        # use Pandas to read columns sets and save transpose
        for i, c in enumerate(column_sets):
            # dtype=str or Pandas converts values into floats
            df = pd.read_csv(input_file, usecols=c, delimiter='\t', dtype=str)
            output_file = join_paths(output_directory, ['chunk{0}.tsv'.format(i)])
            # transpose columns
            df = df.T
            # do not save header that contains row indexes
            df.to_csv(output_file, sep='\t', header=False)
            intermediate_files.append(output_file)

    # concatenate all files with transposed lines
    transposed_file = input_file.replace('.tsv', '_transpose.tsv')
    concatenate_files(intermediate_files, transposed_file)

    # delete intermediate files
    remove_files(intermediate_files)

    return transposed_file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module enables the download of chewBBACA's schemas from the
Chewie-NS.

The process enables the download of ZIP archives that contain ready-to-use
versions of any schema in the Chewie-NS. It also allows users to download
any schema with the structure it had at a specific time point. It is also
possible to download the latest version of the schema through requests to
the Chewie-NS API, if the compressed version that is available does not
match the latest version of the schema. An alternative approach that can
be applied to get the latest version of the schema, if the compressed version
does not provide it, is to download the compressed version that is available
and run the SyncSchema process to retrieve the alleles that were added to the
schema after the compression date.

Expected input
--------------

The process expects the following variables whether through command line
execution or invocation of the :py:func:`main` function:

- ``-sp``, ``species_id`` : The integer identifier or name of the species
  that the schema will be associated to in the Chewie-NS.

    - e.g.: ``1`` or ``'Yersinia pestis'``

- ``-sc``, ``schema_id`` : The schema identifier in the Chewie-NS.

    - e.g.: ``1``

- ``-o``, ``download_folder`` : Path to the parent directory of the folder
  that will store the downloaded schema. The process will create a folder
  with the schema's name inside the directory specified through this argument.

    - e.g.: ``/home/user/chewie_schemas``

- ``--cpu``, ``cpu_cores`` : Number of CPU cores that will be used to
  construct the schema if the process downloads FASTA files instead of
  the compressed version.

    - e.g.: ``4``

- ``--ns_url``, ``nomenclature_server_url`` : The base URL for the Nomenclature
  Server. The default value, "main", will establish a connection to
  "https://chewbbaca.online/", "tutorial" to "https://tutorial.chewbbaca.online/"
  and "local" to "http://127.0.0.1:5000/NS/api/" (localhost). Users may also
  provide the IP address to other Chewie-NS instances.

    - e.g.: ``http://127.0.0.1:5000/NS/api/`` (local host)

- ``--d``, ``date`` : Download schema with structure it had at
  specified date. Must be in the format "Y-m-dTH:M:S" or "Y-m-dTH:M:S.f".

    - e.g.: ``2020-03-27T11:38:00`` or ``2020-03-27T11:38:01.100``

- ``--latest`` : If the compressed version that is available is not the
  latest, downloads all loci and constructs schema locally.

Code documentation
------------------
"""


import os
import sys
import shutil
import requests
import argparse
import concurrent.futures
from itertools import repeat
from urllib3.exceptions import InsecureRequestWarning

try:
    from PrepExternalSchema import PrepExternalSchema
    from utils import (constants as ct,
                       file_operations as fo,
                       process_datetime as pd,
                       chewiens_requests as cr,
                       parameters_validation as pv)
except ModuleNotFoundError:
    from CHEWBBACA.PrepExternalSchema import PrepExternalSchema
    from CHEWBBACA.utils import (constants as ct,
                                 file_operations as fo,
                                 process_datetime as pd,
                                 chewiens_requests as cr,
                                 parameters_validation as pv)


# Suppress only the single warning from urllib3 needed.
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


def check_compressed(schema_uri, headers_get):
    """Determine if there is a compressed version of a schema.

    Parameters
    ----------
    schema_uri : str
        The URI of the schema in the Chewie-NS.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    list
        A list with the following elements:

        - The URI for the compressed version of the schema (str).
        - The timestamp of the compressed version. Indicates the
          last modification date of the schema at time of compression.
    """
    zip_uri, zip_response = cr.simple_get_request(schema_uri,
                                                  headers_get, ['zip'],
                                                  parameters={'request_type': 'check'})
    zip_info = zip_response.json()
    if 'zip' in zip_info:
        zip_file = zip_info['zip'][0]
        zip_date = zip_file.split('_')[-1].split('.zip')[0]
    else:
        zip_date = None

    return [zip_uri, zip_date]


def download_date(user_date, zip_date, latest, insertion_date,
                  modification_date):
    """Determine the date that will be used as timepoint for schema download.

    Parameters
    ----------
    user_date : str or None
        Date value that was received by :py:func:`main`.
        If the user provided a value, it should be a date
        in the format %Y-%m-%dT%H:%M:%S or %Y-%m-%dT%H:%M:%S.%f.
        If the user did not provide a value for the argument, it
        will be None by default.
    zip_date : str or None
        Date value of the compressed version of the schema that
        is available for download. Will be in the format
        %Y-%m-%dT%H:%M:%S or %Y-%m-%dT%H:%M:%S.%f if there is
        a compressed version. Will be None if there is no compressed
        version.
    latest : bool
        Boolean value that indicates if the user wants to download
        the latest version of the schema.
    insertion_date : str
        Date on which the initial schema upload/insertion was
        completed.
    modification_date : str
        Date on which the schema was last modified.

    Returns
    -------
    schema_date : str
        Date selected as timepoint. Schema will be downloaded with
        the data structure it had on this date.

    Raises
    ------
    SystemExit
        - If the value received for the `user_date` is not a
          valid date in the format %Y-%m-%dT%H:%M:%S or
          %Y-%m-%dT%H:%M:%S.%f.
        - If the value received for the `user_date` is a date
          that is prior to `insertion_date` or greater than the
          `modification_date`.
    """
    # user did not provide a value for 'date' or 'latest' arguments
    if user_date is None and latest is False:
        if zip_date is not None:
            schema_date = zip_date
        # there is no compressed version
        else:
            schema_date = modification_date
    # user wants latest schema version
    elif latest is True:
        schema_date = modification_date
    # user wants schema at a particular time point
    elif user_date is not None:
        # get schema insertion and last modification date
        insertion_date_obj = pd.datetime_obj(insertion_date,
                                             '%Y-%m-%dT%H:%M:%S.%f')
        modification_date_obj = pd.datetime_obj(modification_date,
                                                '%Y-%m-%dT%H:%M:%S.%f')
        # determine if date given by user is valid
        user_date_obj = pd.validate_date(user_date)

        if user_date_obj is False:
            sys.exit('Provided date is invalid. Please provide a date '
                     'in the format "%Y-%m-%dT%H:%M:%S" or '
                     '"%Y-%m-%dT%H:%M:%S.%f"')
        if user_date_obj >= insertion_date_obj and \
           user_date_obj <= modification_date_obj:
            schema_date = pd.datetime_str(user_date_obj,
                                          '%Y-%m-%dT%H:%M:%S.%f')
        elif user_date_obj < insertion_date_obj:
            sys.exit('Provided date is prior to the date of schema '
                     'insertion. Please provide a date later than '
                     'schema insertion date ({0}).'.format(insertion_date))
        elif user_date_obj > modification_date_obj:
            sys.exit('Provided date is greater than the last '
                     'modification date. Please provide a date '
                     'equal or prior to the schema modification '
                     'date ({0}).'.format(insertion_date))

    return schema_date


def build_fasta(locus_id, locus_info, download_folder):
    """Write DNA sequences from a response object into a FASTA file.

    Parameters
    ----------
    locus_id : str
        Name of the locus in the Chewie-NS.
    locus_info : dict
        Response object with the list of DNA
        sequences of the locus.
    download_folder : str
        Path to the directory where the FASTA
        file with the locus sequences will be
        created.

    Returns
    -------
    locus_file : str
        Path to the FASTA file with the locus
        sequences.
    """
    locus_name = locus_id.rstrip('.fasta')
    locus_file = os.path.join(download_folder, locus_id+'.fasta')
    ns_data = locus_info.json()['Fasta']

    # extract identifiers and DNA sequences from response
    locus_alleles = []
    for allele in ns_data:
        allele_id = int(allele['allele_id']['value'])
        allele_seq = allele['nucSeq']['value']
        locus_alleles.append((allele_id, allele_seq))

    # create FASTA records
    records = ['>{0}_{1}\n{2}'.format(locus_name, a[0], a[1])
               for a in locus_alleles]

    # write sequences to FASTA file
    with open(locus_file, 'w') as lf:
        concat_records = '\n'.join(records)
        lf.write(concat_records)

    return locus_file


def get_fasta_seqs(url, headers_get, schema_date):
    """Retrieve the DNA sequences of a locus in Chewie-NS.

    Parameters
    ----------
    url : str
        Endpoint URL to make the request.
    headers_get : dict
        HTTP headers for GET requests.
    schema_date : str
        The function will only retrieve alleles
        that were inserted up to this date.

    Returns
    -------
    tuple
        Tuple with the following elements:
        - URI of the locus.
        - Response object with the DNA sequences
          that were downloaded.
    """
    payload = {'date': schema_date}
    tries = 0
    max_tries = 3
    downloaded = False
    while downloaded is False:
        res = cr.simple_get_request(url, headers_get,
                                    [], payload, False, 180)[1]
        tries += 1
        if res.status_code in [200, 201] or tries == max_tries:
            downloaded = True

    return (url.rstrip('/fasta'), res)


def schema_loci(schema_uri, headers_get):
    """Retrieve the list of loci for a schema.

    Parameters
    ----------
    schema_uri : str
        The URI of the schema in the Chewie-NS.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    loci : dict
        A dictionary with loci URIs as keys and
        loci names as values.
    """
    # get the list of loci
    loci_url, loci_res = cr.simple_get_request(schema_uri, headers_get, ['loci'])
    loci_res = loci_res.json()['Loci']

    # locus URI to locus name
    loci = {}
    for locus in loci_res:
        loci[str(locus['locus']['value'])] = locus['name']['value']

    return loci


def download_fastas(loci, download_folder, headers_get, schema_date):
    """Download and write FASTA files for the loci of a schema in Chewie-NS.

    Parameters
    ----------
    loci : dict
        A dictionary with loci URIs as keys and
        loci names as values.
    download_folder : str
        Path to the directory where the FASTA files
        will be created.
    headers_get : dict
        HTTP headers for GET requests.
    schema_date : str
        The function will only retrieve alleles
        that were inserted up to this date.

    Returns
    -------
    ns_files : list
        List with the paths to the schema's FASTA
        files that were created.
    """
    # Total number of loci
    total_loci = len(loci)
    print('Number of loci to download: {0}'.format(total_loci))

    # build the list of urls to get
    fasta_urls = [cr.make_url(locus, 'fasta') for locus in loci]

    # multithread the requests
    print('Downloading schema files...')
    total = 0
    failed = []
    downloaded = 0
    ns_files = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        for result in executor.map(get_fasta_seqs, fasta_urls,
                                   repeat(headers_get), repeat(schema_date)):
            locus_id = loci[result[0]]
            locus_info = result[1]
            if locus_info.status_code in [200, 201]:
                locus_file = build_fasta(locus_id, locus_info, download_folder)
                ns_files.append(locus_file)
                downloaded += 1
            else:
                failed.append(locus_id)
            total += 1
            print('\r', 'Downloaded: '
                  '{0}/{1}'.format(downloaded, total_loci), end='')

    print('\nDownloaded and wrote FASTA files for '
          '{0}/{1} loci'.format(downloaded, total))
    print('Failed download for {0} loci.\n'.format(len(failed)))
    if len(failed) > 0:
        sys.exit('Failed download for following loci: {0}\n'
                 'Please download files for failed loci '
                 'through the API or retry full schema '
                 'download'.format(','.join(failed)))

    return ns_files


def download_compressed(zip_uri, species_name, schema_name,
                        download_folder, headers_get):
    """Download and extract ZIP archive with a ready-to-use schema.

    Parameters
    ----------
    zip_uri : str
        Endpoint URL to make the request to download
        the compressed schema.
    species_name : str
        Scientific name of the schema species.
    schema_name : str
        Name of the schema in the Chewie-NS.
    download_folder : str
        Path to the directory to which the ZIP archive
        will be saved.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    schema_path : str
        ZIP archive contents will be extracted to this
        directory.
    """
    zip_name = '{0}{1}_{2}.zip'.format(species_name[0].lower(),
                                       species_name.split(' ')[-1],
                                       schema_name)
    schema_path = os.path.join(download_folder,
                               zip_name.split('.zip')[0])
    fo.create_directory(schema_path)

    # download ZIP archive
    url, zip_response = cr.simple_get_request(zip_uri, headers_get,
                                              parameters={'request_type': 'download'})
    zip_path = os.path.join(schema_path, zip_name)
    open(zip_path, 'wb').write(zip_response.content)
    # uncompress
    print('Decompressing schema...')
    shutil.unpack_archive(zip_path, extract_dir=schema_path)
    # delete ZIP
    os.remove(zip_path)

    return schema_path


def download_ptf(ptf_hash, download_folder, schema_id,
                 species_id, species_name, headers_get, base_url):
    """Download the Prodigal training file for a schema.

    Parameters
    ----------
    ptf_hash : str
        Unique identifier of the Prodigal training file
        (BLAKE2 hash).
    download_folder : str
        Path to the directory to which the Prodigal
        training file should be saved.
    schema_id : str
        The identifier of the schema in the Chewie-NS.
    species_id : str
        The identifier of the schema's species in the
        Chewie-NS.
    species_name : str
        Scientific name of the schema species.
    headers_get : dict
        HTTP headers for GET requests.
    base_url : str
        Base URL of the Chewie Nomenclature server.

    Returns
    -------
    ptf_file : str
        Path to the Prodigal training file.
    """
    ptf_url, ptf_response = cr.simple_get_request(base_url, headers_get,
                                                  ['species', species_id,
                                                   'schemas', schema_id, 'ptf'])

    ptf_file = os.path.join(download_folder,
                            '{0}.trn'.format(species_name.replace(' ', '_')))

    open(ptf_file, 'wb').write(ptf_response.content)

    return ptf_file


def main(species_id, schema_id, download_folder, cpu_cores,
         nomenclature_server, date, latest, blast_path):

    # GET request headers
    headers_get = ct.HEADERS_GET_JSON

    # Get the name of the species from the provided id
    # or vice-versa
    species_info = cr.species_ids(species_id, nomenclature_server, headers_get)
    if isinstance(species_info, list):
        species_id, species_name = species_info
    else:
        sys.exit('There is no species with the provided '
                 'identifier in the Chewie-NS.')

    # check if user provided schema identifier or schema description
    # get info about all the species schemas
    schema_id, schema_uri,\
        schema_name, schema_params = cr.get_species_schemas(schema_id,
                                                            species_id,
                                                            nomenclature_server,
                                                            headers_get)

    print('Schema id: {0}'.format(schema_id))
    print('Schema name: {0}'.format(schema_name))
    print("Schema's species: {0} "
          "(id={1})".format(species_name, species_id))

    # create parameters dict
    schema_params_dict = {k: schema_params[k]['value']
                          for k in schema_params.keys()
                          if k != 'name'}

    # check if schema is locked
    lock_status = schema_params_dict['Schema_lock']
    if lock_status != 'Unlocked':
        sys.exit('Schema is locked. This might be because it '
                 'is being uploaded, updated or compressed.'
                 ' Please try again later and contact the Administrator '
                 'if the schema stays locked for a long period of time.')

    # get zip information
    zip_uri, zip_date = check_compressed(schema_uri, headers_get)

    schema_date = download_date(date, zip_date, latest,
                                schema_params_dict['dateEntered'],
                                schema_params_dict['last_modified'])

    # check output folder
    if not os.path.exists(download_folder):
        os.mkdir(download_folder)
    else:
        # verify that folder is empty and abort if it is not
        download_folder_files = os.listdir(download_folder)
        if len(download_folder_files) > 0:
            sys.exit('Download folder is not empty. Please ensure '
                     'that folder is empty to guarantee proper '
                     'schema creation or provide a valid path for '
                     'a new folder that will be created.')

    if schema_date == zip_date:
        print('\nDownloading compressed version...')
        # chewie-NS does not add clustering parameters to config, change that
        schema_path = download_compressed(zip_uri, species_name, schema_name,
                                          download_folder, headers_get)
    else:
        print('\nDownloading schema FASTA files...')
        # download FASTA files
        loci = schema_loci(schema_uri, headers_get)
        ns_files = download_fastas(loci, download_folder, headers_get,
                                   schema_date)

        # download Prodigal training file
        ptf_hash = schema_params_dict['prodigal_training_file']
        ptf_file = download_ptf(ptf_hash, download_folder, schema_id,
                                species_id, species_name, headers_get,
                                nomenclature_server)

        # use PrepExternalSchema main to determine representatives
        genus, epithet = species_name.split(' ')
        schema_name = '{0}{1}_{2}'.format(genus[0].lower(), epithet, schema_name)
        schema_path = os.path.join(download_folder, schema_name)

        # determine representatives and create schema
        PrepExternalSchema.main(download_folder,
                                schema_path,
                                cpu_cores,
                                float(schema_params_dict['bsr']),
                                int(schema_params_dict['minimum_locus_length']),
                                int(schema_params_dict['translation_table']),
                                ptf_file,
                                None,
                                blast_path)

        # copy Prodigal training file to schema directory
        shutil.copy(ptf_file, schema_path)
        os.remove(ptf_file)

        # remove FASTA files with sequences from the NS
        fo.remove_files(ns_files)

        # write hidden schema config file
        del(schema_params_dict['Schema_lock'])
        schema_config = pv.write_schema_config(schema_params_dict['bsr'],
                                               ptf_hash,
                                               schema_params_dict['translation_table'],
                                               schema_params_dict['minimum_locus_length'],
                                               schema_params_dict['chewBBACA_version'],
                                               schema_params_dict['size_threshold'],
                                               schema_params_dict['word_size'],
                                               schema_params_dict.get('window_size', None),  # add window size to chewie-NS
                                               schema_params_dict['cluster_sim'],
                                               schema_params_dict['representative_filter'],
                                               schema_params_dict['intraCluster_filter'],
                                               schema_path)

        # create ns_config file
        ns_config = os.path.join(schema_path, '.ns_config')
        download_info = [schema_date, schema_uri]
        if not os.path.exists(ns_config):
            fo.pickle_dumper(download_info, ns_config)

        genes_list_file = pv.write_gene_list(schema_path)

    print('Schema is now available at: {0}'.format(schema_path))


def parse_arguments():

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-sp', type=str, required=True,
                        dest='species_id',
                        help='The integer identifier or name of the '
                             'species that the schema is associated '
                             'to in the Chewie-NS.')

    parser.add_argument('-sc', type=str, required=True,
                        dest='schema_id',
                        help='The schema identifier in the Chewie-NS.')

    parser.add_argument('-o', type=str, required=True,
                        dest='download_folder',
                        help='Path to the parent directory of the '
                             'folder that will store the downloaded schema.')

    parser.add_argument('--cpu', type=int, required=False,
                        dest='cpu_cores', default=1,
                        help='Number of CPU cores that will '
                             'be used to construct the schema '
                             'if the process downloads FASTA '
                             'files instead of a compressed version.')

    parser.add_argument('--ns', type=pv.validate_ns_url, required=False,
                        dest='nomenclature_server',
                        default='main',
                        help='The base URL for the Nomenclature Server. '
                             'The default value, "main", will establish a '
                             'connection to "https://chewbbaca.online/", '
                             '"tutorial" to "https://tutorial.chewbbaca.online/" '
                             'and "local" to "http://127.0.0.1:5000/NS/api/" (localhost). '
                             'Users may also provide the IP address to other '
                             'Chewie-NS instances.')

    parser.add_argument('--d', type=str, required=False,
                        default=None,
                        dest='date',
                        help='Download schema with structure it had at '
                             'specified date. Must be in the format '
                             '"Y-m-dTH:M:S" or "Y-m-dTH:M:S.f".')

    parser.add_argument('--latest', required=False,
                        action='store_true', dest='latest',
                        help='If the compressed version that is available '
                             'is not the latest, downloads all loci and '
                             'constructs schema locally.')

    parser.add_argument('--b', type=pv.check_blast, required=False,
                        default='', dest='blast_path',
                        help='Path to the BLAST executables.')

    args = parser.parse_args()

    return args


if __name__ == "__main__":

    args = parse_arguments()
    main(**vars(args))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions related with sequence clustering
based on k-mer sets.

Code documentation
------------------
"""


import os
from collections import Counter

try:
    from utils import (file_operations as fo,
                       iterables_manipulation as im,
                       blast_wrapper as bw,
                       fasta_operations as fao,
                       constants as ct)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (file_operations as fo,
                                 iterables_manipulation as im,
                                 blast_wrapper as bw,
                                 fasta_operations as fao,
                                 constants as ct)


def select_representatives(kmers, reps_groups, clustering_sim):
    """Determine the clusters a sequence can be added to.

    Determines the set of clusters that a sequence can be
    added to based on the decimal proportion of shared
    distinct k-mers.

    Parameters
    ----------
    kmers : list or set
        Set of k-mers determined by decomposing a single
        sequence.
    reps_groups : dict
        Dictionary with k-mers as keys and sequence
        identifiers of sequences that contain each
        k-mer as values.
    clustering_sim : float
        Sequences are added to clusters if they
        share a minimum decimal proportion of
        distinct k-mers with a cluster representative.

    Returns
    -------
    selected_reps : list
        List with a tuple per cluster/representative
        that the sequence can be added to. Each tuple
        has the identifier of the cluster representative
        and the decimal proportion of shared distinct
        kmers.
    """
    current_reps = [reps_groups[k] for k in kmers if k in reps_groups]
    current_reps = im.flatten_list(current_reps)

    # count number of kmer hits per representative
    counts = Counter(current_reps)
    selected_reps = [(k, v/len(kmers))
                     for k, v in counts.items()
                     if v/len(kmers) >= clustering_sim]

    # sort by identifier and then by similarity to always get same order
    selected_reps = sorted(selected_reps, key=lambda x: x[0])
    selected_reps = sorted(selected_reps, key=lambda x: x[1], reverse=True)

    return selected_reps


def intra_cluster_sim(clusters, sequences, word_size, intra_filter):
    """Identify clustered sequences that are similar.

    Determines the percentage of shared k-mers/minimizers
    between sequences in the same cluster and excludes
    sequences that are similar to other sequences in the
    cluster.

    Parameters
    ----------
    clusters : dict
        Dictionary with the identifiers of sequences
        that are cluster representatives as keys and
        a list with tuples as values. Each tuple has
        the identifier of a sequence that was added to
        the cluster, the percentage of shared
        kmers/minimizers and the length of the clustered
        sequence.
    sequences : dict
        Dictionary with sequence identifiers as keys
        and sequences as values.
    word_size : int
        Value k for the kmer size.
    intra_filter : float
        Similarity threshold value. If two sequences in
        the same cluster have a similarity value equal
        or greater to this value, the shorter sequence
        will be excluded from the cluster.

    Returns
    -------
    excluded_seqids : dict
        Dictionary with the identifiers of sequences
        that are cluster representatives as keys and
        a list with the sequence identifiers of sequences
        that were excluded from the cluster as values.
    excluded_sims : dict
        A dictionary with the identifiers of sequences
        that are cluster representatives as keys
        and a list with tuples as values. Each tuple
        contains the sequence identifier and the
        similarity value for a match that led to
        an exclusion.
    """
    excluded_seqids = {}
    excluded_sims = {}
    for representative, clustered in clusters.items():
        # get identifiers of sequences in the cluster
        clustered_ids = [c[0] for c in clustered]
        clustered_seqs = {seqid: sequences[seqid] for seqid in clustered_ids}

        # create kmer index
        kmers_mapping = im.kmer_index(clustered_seqs, word_size, fasta=False)

        excluded = []
        similarities = []
        # for each sequence in the cluster
        for seqid, sequence in clustered_seqs.items():
            if seqid not in excluded:
                query_kmers = im.determine_minimizers(sequence, word_size,
                                                      word_size, position=False)

                # select sequences with same kmers
                sims = select_representatives(query_kmers,
                                              kmers_mapping,
                                              intra_filter)

                if len(sims) > 1:
                    # exclude current sequence
                    candidates = [s for s in sims if s[0] != seqid]
                    for c in candidates:
                        # if query sequence is longer, keep it and exclude candidate
                        if len(clustered_seqs[seqid]) >= len(clustered_seqs[c[0]]):
                            picked = (c[0], seqid, c[1])
                        # otherwise, exclude query sequence
                        elif len(clustered_seqs[c[0]]) > len(clustered_seqs[seqid]):
                            picked = (seqid, c[0], c[1])
                        similarities.append(picked)
                        excluded.append(picked[0])

        # convert into set first to remove possible duplicates
        excluded_seqids[representative] = list(set(excluded))
        excluded_sims[representative] = similarities

    return [excluded_seqids, excluded_sims]


# change name and add parameter to accept k-mer sampling used to cluster?
def minimizer_clustering(sorted_sequences, word_size, window_size, position,
                         offset, clusters, reps_sequences, reps_groups,
                         seq_num_cluster, clustering_sim, grow):
    """Cluster sequences based on shared distinct minimizers.

    Parameters
    ----------
    sorted_sequences : dict
        Dictionary with sequence identifiers as keys and
        sequences as values. Sorted by decreasing sequence
        length.
    word_size : int
        Value k for the kmer size.
    window_size : int
        Window size used to determine minimizers.
    position : bool
        If minimizer sequence position should be stored.
    offset : int
        Value to indicate offset of consecutive kmers.
    clusters : dict
        Dictionary with the identifiers of sequences
        that are cluster representatives as keys and
        a list with tuples as values. Each tuple has
        the identifier of a sequence that was added to
        the cluster, the decimal proportion of shared
        distinct minimizers and the length of the clustered
        sequence. This dictionary should be empty at the
        start of the clustering step during the CreateSchema
        process. For the AlleleCall process, the dictionary
        should contain the identifiers of the loci
        representatives.
    reps_sequences : dict
        Dictionary with the identifiers of sequences
        that are cluster representatives as keys and
        their sequences as values. This dictionary should
        be empty at the start of the clustering step
        during the CreateSchema process. For the AlleleCall
        process, the dictionary should contain the
        identifiers of the loci representatives as keys
        and their sequences as values.
    reps_groups : dict
        Dictionary with kmers as keys and a list with
        identifiers of sequences that contain that kmer
        as values. This dictionary should be empty at the
        start of the clustering step during the CreateSchema
        process. For the AlleleCall process, the dictionary
        should contain all kmers contained in the set of
        the schema's representatives sequences.
    seq_num_cluster : int
        Maximum number of clusters that a sequence can be
        added to.
    clustering_sim : float
        Similarity threshold to cluster a sequence into
        a cluster.

    Returns
    -------
    A list with the following elements:
        clusters : dict
            Dictionary with the identifiers of sequences
            that are clusters representatives as keys and
            a list with tuples as values. Each tuple has
            the identifier of a sequence that was added to
            the cluster, the decimal proportion of shared
            distinct minimizers and the length of the clustered
            sequence.
        reps_sequences : dict
            Dictionary with the identifiers of sequences
            that are cluster representatives as keys and
            their sequences as values.
        reps_groups : dict
            Dictionary with kmers as keys and a list with
            identifiers of sequences that contain that kmer
            as values.
    """
    for protid, protein in sorted_sequences.items():

        minimizers = im.determine_minimizers(protein, window_size,
                                             word_size, offset=offset,
                                             position=position)

        distinct_minimizers = set(minimizers)

        selected_reps = select_representatives(distinct_minimizers,
                                               reps_groups,
                                               clustering_sim)

        top = (len(selected_reps)
               if len(selected_reps) < seq_num_cluster
               else seq_num_cluster)

        # sort to get most similar at index 0
        if len(selected_reps) > 0:
            for i in range(0, top):
                clusters[selected_reps[i][0]].append((protid,
                                                      selected_reps[i][1],
                                                      len(protein),
                                                      len(minimizers),
                                                      len(distinct_minimizers)))
        else:
            if grow is True:
                for k in distinct_minimizers:
                    reps_groups.setdefault(k, []).append(protid)

                clusters[protid] = [(protid, 1.0, len(protein),
                                    len(minimizers), len(distinct_minimizers))]
                reps_sequences[protid] = protein

    return [clusters, reps_sequences, reps_groups]


def clusterer(sorted_sequences, word_size, window_size,
              clustering_sim, representatives, grow,
              offset, position, seq_num_cluster):
    """Cluster sequences based on shared distinct minimizers.

    Parameters
    ----------
    sorted_sequences : dict
        Dictionary with sequence identifiers as keys and
        sequences as values. Sorted by decreasing sequence
        length.
    word_size : int
        Value k for the kmer size.
    window_size : int
        Window size used to determine minimizers.
    clustering_sim : float
        Similarity threshold to cluster a sequence into
        a cluster.
    representatives : dict
        Dictionary with kmers as keys and a list with
        identifiers of sequences that contain that kmer
        as values.
    grow : bool
        If it is allowed to create new clusters.
    offset : int
        Value to indicate offset of consecutive kmers.
    position : bool
        If minimizer sequence position should be stored.
    seq_num_cluster : int
        Maximum number of clusters that a sequence can be
        added to.

    Returns
    -------
    A list with the following elements:
        clusters : dict
            Dictionary with the identifiers of sequences
            that are cluster representatives as keys and
            a list with tuples as values. Each tuple has
            the identifier of a sequence that was added to
            the cluster, the decimal proportion of shared
            distinct minimizers and the length of the clustered
            sequence.
        reps_sequences : dict
            Dictionary with the identifiers of sequences
            that are cluster representatives as keys and
            their sequences as values.
    """
    clusters = {}
    reps_sequences = {}
    if representatives is None:
        reps_groups = {}
    else:
        reps_groups = representatives
        clusters_ids = set(im.flatten_list(list(reps_groups.values())))
        clusters = {rep: [] for rep in clusters_ids}

    cluster_results = minimizer_clustering(sorted_sequences, word_size,
                                           window_size, position,
                                           offset, clusters,
                                           reps_sequences, reps_groups,
                                           seq_num_cluster, clustering_sim,
                                           grow)

    return cluster_results[0:2]


def write_clusters(clusters, outfile):
    """Write information about clusters to file.

    Parameters
    ----------
    clusters : dict
        Dictionary with the identifiers of sequences
        that are cluster representatives as keys and
        a list with tuples as values. Each tuple has
        the identifier of a sequence that was added to
        the cluster, the decimal proportion of shared
        distinct kmers/minimizers and the length of the
        clustered sequence.
    outfile : str
        Path to the file that will be created to save
        information about clusters.
    """
    cluster_lines = []
    for rep, seqids in clusters.items():
        current_cluster = []
        current_cluster.append('>{0}'.format(rep))
        clustered = [', '.join(['{}']*len(s)).format(*s)
                     for s in seqids]
        current_cluster.extend(clustered)
        cluster_lines.append(current_cluster)

    # sort by number of lines to get clusters with more sequences first
    cluster_lines = im.sort_iterable(cluster_lines,
                                     sort_key=lambda x: len(x), reverse=True)
    cluster_lines = im.flatten_list(cluster_lines)
    cluster_text = im.join_list(cluster_lines, '\n')

    fo.write_to_file(cluster_text, outfile, 'w', '\n')


def representative_pruner(clusters, sim_cutoff):
    """Remove sequences from clusters based on a similarity threshold.

    Parameters
    ----------
    clusters : dict
        Dictionary with the identifiers of sequences
        that are cluster representatives as keys and
        a list with tuples as values. Each tuple has
        the identifier of a sequence that was added to
        the cluster, the percentage of shared
        kmers/minimizers and the length of the clustered
        sequence.
    sim_cutoff : float
        Similarity threshold value. Sequences with
        equal or greater similarity value are excluded
        from clusters.

    Returns
    -------
    A list with the following elements:
        pruned_clusters : dict
            Input dictionary without values for the
            sequences that had similarity values
            equal or greater that defined threshold.
        excluded : list
            List with a list per cluster. Each list
            has the sequences that were excluded from
            a cluster.
    """
    excluded = []
    pruned_clusters = {}
    for rep, seqids in clusters.items():
        pruned_clusters[rep] = [seqid
                                for seqid in seqids
                                if seqid[1] < sim_cutoff]
        # do not exclude cluster representatives
        excluded.extend([seqid
                         for seqid in seqids
                         if seqid[1] >= sim_cutoff and seqid[0] != rep])

    return [pruned_clusters, excluded]


def cluster_blaster(seqids, sequences, output_directory,
                    blast_path, blastdb_path, only_rep=False):
    """Align sequences in the same cluster with BLAST.

    Parameters
    ----------
    seqids : list
        List with cluster identifiers.
    sequences : str
        Path to the FASTA file with the protein sequences
        in all clusters.
    output_directory : str
        Path to the directory where FASTA files with
        the sequences in each cluster and files with
        BLAST results will be written to.
    blast_path : str
        Path to BLAST executables
    blastdb_path : str
        Path to a BLAST database.

    Returns
    -------
    out_files : list
        List with the paths to the files with the BLAST
        results for each cluster.
    """
    # have to index file here, cannot provide index as argument while multiprocessing, why?
    indexed_fasta = fao.index_fasta(sequences)

    out_files = []
    for cluster in seqids:

        cluster_id = cluster
        ids_file = os.path.join(output_directory,
                                '{0}_ids.txt'.format(cluster_id))

        fasta_file = os.path.join(output_directory,
                                  '{0}_protein.fasta'.format(cluster_id))

        if only_rep is False:
            with open(ids_file, 'r') as clstr:
                cluster_ids = [l.strip() for l in clstr.readlines()]
            # create file with protein sequences
            fao.get_sequences_by_id(indexed_fasta, cluster_ids, fasta_file)
        else:
            fao.get_sequences_by_id(indexed_fasta, [cluster_id], fasta_file)

        blast_output = os.path.join(output_directory,
                                    '{0}_blast_out.tsv'.format(cluster_id))

        # Use subprocess to capture errors and warnings
        stderr = bw.run_blast(blast_path, blastdb_path, fasta_file,
                              blast_output, 1, 1,
                              ids_file, ignore=ct.IGNORE_RAISED)

        if len(stderr) > 0:
            raise ValueError('\n'.join(stderr))

        out_files.append(blast_output)

    return out_files


def blast_seqids(clusters, output_directory, ids_dict, only_rep):
    """Create files with the identifiers of the sequences in each cluster.

    Parameters
    ----------
    clusters : dict
        Dictionary with the identifiers of cluster
        representatives as keys and a list with tuples
        as values (each tuple has the identifier of a
        sequence that is in the cluster, the decimal
        proportion of shared minimizers and the length
        of that sequence).
    output_directory : str
        Path to the directory where files with identifiers
        will be created.
    ids_dict : dict
        Dictionary that maps sequence identifiers to
        shorter and unique identifiers that will be
        saved in the files and used as sequence
        identifiers during BLAST to avoid errors
        related with sequence headers/identifiers
        that exceed length limit allowed by BLAST.

    Returns
    -------
    ids_to_blast : list
        List with the identifiers of all clusters.
    """
    rev_ids = {v: k for k, v in ids_dict.items()}

    ids_to_blast = []
    for rep in clusters:

        cluster_file = os.path.join(output_directory,
                                    '{0}_ids.txt'.format(rev_ids[rep]))
        if only_rep is False:
            cluster_ids = [rev_ids[rep]] + [rev_ids[seqid[0]]
                                            for seqid in clusters[rep]]
        else:
            cluster_ids = [rev_ids[seqid[0]] for seqid in clusters[rep]]

        cluster_lines = im.join_list(cluster_ids, '\n')
        fo.write_to_file(cluster_lines, cluster_file, 'w', '')
        ids_to_blast.append((rev_ids[rep], len(cluster_ids)))

    return ids_to_blast



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This modules contains functions used to paralellize
function calls.

Code documentation
------------------
"""


import time
import traceback
from multiprocessing import Pool

try:
    from utils import iterables_manipulation as im
except ModuleNotFoundError:
    from CHEWBBACA.utils import iterables_manipulation as im


def function_helper(input_args):
    """Run function with provided inputs and capture exceptions.

    Parameters
    ----------
    input_args : list
        List with function inputs and function object to call
        in the last index.

    Returns
    -------
    results : list
        List with the results returned by the function.
        If an exception is raised it returns a list with
        the name of the function and the exception traceback.
    """
    try:
        results = input_args[-1](*input_args[0:-1])
    except Exception as e:
        func_name = (input_args[-1]).__name__
        traceback_lines = traceback.format_exception(etype=type(e), value=e,
                                                     tb=e.__traceback__)
        traceback_text = ''.join(traceback_lines)
        print('\nError on {0}:\n{1}\n'.format(func_name, traceback_text))
        results = [func_name, traceback_text]

    return results


def map_async_parallelizer(inputs, function, cpu, callback='extend',
                           chunksize=1, show_progress=False):
    """Run function in parallel.

    Parameters
    ----------
    inputs : list
        List with inputs to process.
    function : func
        Function to be parallelized.
    cpu : int
        Number of processes to create (based on the
        number of CPU cores).
    callback : str
        Results can be appended, "append", to the
        list that stores results or the list of results
        can be extended, "extend".
    chunksize : int
        Size of input chunks that will be passed to
        each process. The function will create groups
        of inputs with this number of elements.
    show_progress: bool
        True to show a progress bar with the percentage
        of inputs that have been processed, False
        otherwise.

    Returns
    -------
    results : list
        List with the results returned for each function
        call.
    """
    results = []
    # use context manager to join and close pool automatically
    with Pool(cpu) as pool:
        if callback == 'extend':
            rawr = pool.map_async(function, inputs,
                                  callback=results.extend, chunksize=chunksize)
        elif callback == 'append':
            rawr = pool.map_async(function, inputs,
                                  callback=results.append, chunksize=chunksize)

        if show_progress is True:
            progress = None
            while progress != 100:
                progress = progress_bar(rawr._number_left, len(inputs), progress)

        rawr.wait()

    return results


def progress_bar(remaining, total, previous, tickval=5, ticknum=20):
    """Create and print a progress bar to the stdout.

    Parameters
    ----------
    remaining : int
        Number of remaining tasks to complete.
    total : int
        Total number of inputs that have to be processed.
    previous : int
        Percentage of tasks that had been completed in the
        previous function call.
    tickval : int
        Progress completion percentage value for each
        tick.
    ticknum : int
        Total number of ticks in progress bar.

    Returns
    -------
    completed : bool
        Boolean indicating if all inputs have been processed.
    """
    # determine percentage of processed inputs
    progress = int(100-(remaining/total)*100)
    # only print if percentage has changed
    if progress != previous:
        progress_tick = progress//tickval
        progress_bar = '[{0}{1}] {2}%'.format('='*progress_tick,
                                              ' '*(ticknum-progress_tick),
                                              progress)
        print('\r', progress_bar, end='')

    time.sleep(0.1)

    return progress


def distribute_loci(inputs, cores, method):
    """Create balanced lists of loci to efficiently parallelize function calls.

    Creates balanced lists of loci to distribute per number of
    available cores. Loci lists can be created based on the number
    of sequences per locus (seqcount), the mean length of the
    sequences (length) in each locus or the product of both values
    (seqcount+length).

    Parameters
    ----------
    inputs : list
        List with one sublist per locus. Each sublist has
        a locus identifier, the total number of sequences
        and sequence mean legth for that locus.
    cores : int
        The number of loci groups that should be created.
        Based on the number of CPU cores that will be
        used to process the inputs.
    method : str
        "seqcount" to create loci lists based on the total
        number of sequences, "length" to split based
        on mean length of sequences and "seqcount+length" to
        split based on both criteria.

    Returns
    -------
    splitted_ids : list
        List with sublists that contain loci identifiers.
        Sublists are balanced based on the chosen method.
    """
    # initialize list with sublists to store inputs
    splitted_ids = [[] for cpu in range(cores)]
    # initialize list with chosen criterion values
    # for each sublist of inputs
    splitted_values = [0 for cpu in range(cores)]
    i = 0
    for locus in inputs:
        if method == 'seqcount':
            splitted_values[i] += locus[1]
        elif method == 'length':
            splitted_values[i] += locus[4]
        elif method == 'seqcount+length':
            splitted_values[i] += locus[1] * locus[4]
        splitted_ids[i].append(locus[0])
        # at the end of each iteration, choose the sublist
        # with lowest criterion value
        i = splitted_values.index(min(splitted_values))

    return splitted_ids


def parallelize_function(function, inputs, common_args=None,
                         cpu_cores=1, show_progress=False):
    """Create list of inputs and parallelize function calls.

    Parameters
    ----------
    function : func
        Function to be parallelized.
    inputs : list
        List of inputs to divide into sublists.
    common_args : list
        List of common arguments to add to each sublist.
        The common args are values that will be passed to the
        function.
    cpu_cores : int
        Number of CPU cores used to parallelize the function.
    show_progress : bool
        True to show a progress bar for the percentage of
        inputs that have been processed, False otherwise.

    Returns
    -------
    results : list
        List with the results returned by the function calls.
    """
    # create chunks to distribute per cores
    input_lists = im.divide_list_into_n_chunks(inputs, len(inputs))

    if common_args is None:
        common_args = []

    # add common arguments to all sublists
    input_lists = im.multiprocessing_inputs(input_lists,
                                            common_args,
                                            function)

    results = map_async_parallelizer(input_lists,
                                     function_helper,
                                     cpu_cores,
                                     show_progress=show_progress)

    return results



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions related with biological sequence
manipulation.

Code documentation
------------------
"""


from Bio.Seq import Seq
from collections import Counter

try:
    from utils import (constants as ct,
                       file_operations as fo,
                       iterables_manipulation as im,
                       fasta_operations as fao)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 file_operations as fo,
                                 iterables_manipulation as im,
                                 fasta_operations as fao)


def translate_sequence(dna_str, table_id):
    """Translate a DNA sequence using the BioPython package.

    Parameters
    ----------
    dna_str : str
        String representing a DNA sequence.
    table_id : int
        Translation table identifier.

    Returns
    -------
    protseq : Bio.Seq.Seq
        Protein sequence created by translating the
        input DNA sequence.
    """
    myseq_obj = Seq(dna_str)
    protseq = Seq.translate(myseq_obj, table=table_id, cds=True)

    return protseq


def translate_dna_aux(dna_sequence, method, table_id):
    """Attempt to translate a DNA sequence in specified orientation.

    Attempts to translate an input DNA sequence in specified
    orientation and stores exceptions when the input sequence
    cannot be translated.

    Parameters
    ----------
    dna_sequence : str
        String representing a DNA sequence.
    method : str
        Sequence orientation to attempt translation.
    table_id : int
        Translation table identifier.

    Returns
    -------
    If the sequence can be translated:
        protseq : Bio.Seq.Seq
            Translated DNA sequence.
        myseq : str
            String representing the DNA sequence in the
            orientation used to translate it.
    Otherwise, returns string with the description of the
    exception that was raised.
    """
    myseq = dna_sequence
    # try to translate original sequence
    if method == 'original':
        try:
            protseq = translate_sequence(myseq, table_id)
        except Exception as argh:
            return argh
    # try to translate the reverse complement
    elif method == 'revcomp':
        try:
            myseq = im.reverse_complement(myseq, ct.DNA_BASES)
            protseq = translate_sequence(myseq, table_id)
        except Exception as argh:
            return argh
    # try to translate the reverse
    elif method == 'rev':
        try:
            myseq = im.reverse_str(myseq)
            protseq = translate_sequence(myseq, table_id)
        except Exception as argh:
            return argh
    # try to translate the reverse reverse complement
    elif method == 'revrevcomp':
        try:
            myseq = im.reverse_str(myseq)
            myseq = im.reverse_complement(myseq, ct.DNA_BASES)
            protseq = translate_sequence(myseq, table_id)
        except Exception as argh:
            return argh

    return [protseq, myseq]


def translate_dna(dna_sequence, table_id, min_len):
    """Determine if a DNA sequence is valid and translate it.

    Checks if sequence is valid and attempts to translate
    it, calling several functions to ensure that the sequence
    only has 'ACTG', is multiple of 3 and that it can be
    translated in any of 4 different orientations. Stores
    exceptions so that it is possible to understand why the
    sequence could not be translated.

    Parameters
    ----------
    dna_sequence : str
        String representing a DNA sequence.
    table_id : int
        Translation table identifier.
    min_len : int
        Minimum sequence length. Sequences shorter
        than this value are not translated.

    Returns
    -------
    If the sequence can be translated:
        sequence : list
            List with two elemets, the protein sequence
            and the DNA sequence in the correct orientation.
        coding_strand : str
            The sequence orientation that codes for the
            protein.
    Otherwise:
        exception_str : str
            A string containing the exceptions that
            explain why the the sequence could not be
            translated.
    """
    original_seq = dna_sequence.upper()
    exception_collector = []
    strands = ['sense', 'antisense', 'revsense', 'revantisense']
    translating_methods = ['original', 'revcomp', 'rev', 'revrevcomp']

    # check if the sequence has ambiguous bases
    valid_dna = im.check_str_alphabet(original_seq, ct.DNA_BASES)
    if valid_dna is not True:
        return 'ambiguous or invalid characters'

    # check if sequence size is multiple of three
    valid_length = im.check_str_multiple(original_seq, 3)
    if valid_length is not True:
        return 'sequence length is not a multiple of 3'

    # check if sequence is not shorter than the accepted minimum length
    if len(original_seq) < min_len:
        return 'sequence shorter than {0} nucleotides'.format(min_len)

    # try to translate in 4 different orientations
    # or reach the conclusion that the sequence cannot be translated
    i = 0
    translated = False
    while translated is False:
        translated_seq = translate_dna_aux(original_seq, translating_methods[i], table_id)
        if not isinstance(translated_seq, list):
            exception_collector.append('{0}({1})'.format(strands[i],
                                                         translated_seq.args[0]))

        i += 1
        if i == len(strands) or isinstance(translated_seq, list) is True:
            translated = True

    coding_strand = strands[i-1]

    # if the sequence could be translated, return list with protein and DNA
    # sequence in correct orientation
    if isinstance(translated_seq, list):
        return [translated_seq, coding_strand]
    # if it could not be translated, return the string with all exception
    # that were collected
    else:
        exception_str = ','.join(exception_collector)
        return exception_str


def determine_duplicated_seqs(sequences):
    """Create mapping between sequences and sequence identifiers.

    Parameters
    ----------
    sequences : dict
        Dictionary with sequence identifiers as keys and
        sequences as values.

    Returns
    -------
    equal_seqs : dict
        Dictionary with sequences as keys and sequence
        identifiers that are associated with each
        sequence as values.
    """
    equal_seqs = {}
    for seqid, seq in sequences.items():
        # if protein sequence was already added as key
        if seq in equal_seqs:
            # append new protid
            equal_seqs[seq].append(seqid)
        # else add new protein sequence as key and protid
        # as value
        else:
            equal_seqs[seq] = [seqid]

    return equal_seqs


def determine_longest(seqids, sequences):
    """Find the longest sequence in a set of sequences.

    Parameters
    ----------
    seqids : list
        List with sequence identifiers.
    sequences : dict
        Dictionary with sequence identifiers as keys
        and sequences as values.

    Returns
    -------
    chosen : str
        Sequence identifier of the longest sequence.
    """
    seqids_tups = [(seqid, sequences[seqid]) for seqid in seqids]
    sorted_tups = sorted(seqids_tups, key=lambda x: len(x[1]), reverse=True)
    chosen = sorted_tups[0][0]

    return chosen


def determine_mode(values):
    """Compute the mode based on a list of integers.

    Parameters
    ----------
    values : list
        List with integer values.

    Returns
    -------
    modes : list
        The most frequent integer values.
    """
    # determine frequency of each length value
    counts = Counter(values)

    # order by most common first
    most_common = counts.most_common()

    # get most common
    modes = [most_common[0][0]]

    # determine if there are more length values that are just as common
    modes += [m[0] for m in most_common[1:] if m[1] == most_common[0][1]]

    return modes


def mode_filter(sequences, size_threshold):
    """Find sequences with length that deviates from the mode value.

    Determines the mode from a set of input sequences and
    identifies sequences that have a length value smaller
    or greater than the mode based on a threshold.

    Parameters
    ----------
    sequences : dict
        Dictionary with sequence identifiers as keys and
        sequences as values.
    size_threshold : float
        Sequences with +/- this value * mode will be
        reported as above or below the mode.

    Returns
    -------
    A list with the following variables:
        modes : list
            List with mode values determined based on the
            length of input sequences.
        alm : list
            List with the sequence identifiers of the
            sequences that are above the mode value by
            mode*size_threshold.
        asm : list
            List with the sequence identifiers of the
            sequences that are below the mode value by
            mode*size_threshold.
        seqs_lengths : dict
            Dictionary with sequence identifiers as keys
            and sequence lengths as values.
    """
    # determine length value of all sequences
    seqs_lengths = {seqid: len(seq) for seqid, seq in sequences.items()}

    # determine mode/s
    modes = determine_mode(list(seqs_lengths.values()))

    # determine top and bot length value limits
    max_mode = max(modes)
    top_limit = max_mode + (max_mode*size_threshold)
    min_mode = min(modes)
    bot_limit = min_mode - (min_mode*size_threshold)

    # determine sequences that are below or above limits
    alm = [seqid for seqid, length in seqs_lengths.items()
           if length > top_limit]
    asm = [seqid for seqid, length in seqs_lengths.items()
           if length < bot_limit]

    return [modes, alm, asm, seqs_lengths]


def get_seqs_dicts(fasta_path, gene_id, table_id, min_len, size_threshold):
    """Translate DNA sequences in a FASTA file.

    Translates the set of alleles from a gene. Identifies
    sequences that cannot be translated according to the
    criteria enforced by chewBBACA.

    Parameters
    ----------
    fasta_path : str
        Path to the FASTA file with DNA sequences.
    gene_id : str
        Gene identifier.
    table_id : int
        Translation table identifier.
    min_len : int
        Minimum sequence length. Sequences shorter
        than this value are not translated.
    size_threshold : float
        Sequences with +/- this value * mode will be
        reported as above or below the mode.

    Returns
    -------
    List with following elements:
        dna_seqs : dict
            Dictionary with sequence identifiers as keys
            and DNA sequences as values.
        prot_seqs : dict
            Dictionary with protein identifiers as keys
            and Protein sequences as values. Keys are
            consecutive integers to enable alignment
            with BLASTp without getting exceptions due
            to long sequence identifiers.
        invalid : list
            List with sequence identifiers of alleles
            that are not valid because they could not be
            translated.
        seqids_map : dict
            Dictionary with consecutive integers as keys
            and original allele identifiers as values.
        total_seqs : int
            Total number of sequences that was processed
            (including invalid alleles).
    """
    sequences = fao.import_sequences(fasta_path)

    # translate sequences
    translated_seqs = {k: translate_dna(v, table_id, min_len)
                       for k, v in sequences.items()}

    # add locus identifier to headers
    # some headers might only have the allele identifier
    seqids = list(translated_seqs.keys())
    new_seqids = {}
    for i in seqids:
        new_id = '{0}_{1}'.format(gene_id, i.split('_')[-1])
        new_seqids[i] = new_id

    # switch ids
    sequences = {new_seqids[k]: v
                 for k, v in translated_seqs.items()}

    valid = {k: v
             for k, v in sequences.items()
             if isinstance(v, list) is True}
    invalid = [[k, v]
               for k, v in sequences.items()
               if isinstance(v, list) is False]

    seqid = 1
    seqids_map = {}
    dna_seqs = {}
    prot_seqs = {}
    for k, v in valid.items():
        seqids_map[str(seqid)] = k
        dna_seqs[k] = v[0][1]
        prot_seqs[str(seqid)] = str(v[0][0])
        seqid += 1

    if size_threshold is not None and len(prot_seqs) > 0:
        # remove alleles based on length mode and size threshold
        modes, alm, asm, alleles_lengths = mode_filter(dna_seqs, size_threshold)
        excluded = set(asm + alm)

        dna_seqs = {seqid: seq
                    for seqid, seq in dna_seqs.items()
                    if seqid not in excluded}
        prot_seqs = {seqid: seq
                     for seqid, seq in prot_seqs.items()
                     if seqids_map[seqid] not in excluded}

        modes_concat = ':'.join(map(str, modes))
        st_percentage = int(size_threshold*100)
        invalid += [[s, 'allele greater than {0}% locus length mode '
                     '({1}>{2})'.format(st_percentage,
                                        alleles_lengths[s],
                                        modes_concat)]
                    for s in alm]
        invalid += [[s, 'allele smaller than {0}% locus length mode '
                     '({1}<{2})'.format(st_percentage,
                                        alleles_lengths[s],
                                        modes_concat)]
                    for s in asm]

    total_seqs = len(translated_seqs)

    return [dna_seqs, prot_seqs, invalid, seqids_map, total_seqs]


def translate_coding_sequences(seqids, protein_file, sequences_file,
                               translation_table, minimum_length):
    """Translate coding sequences.

    Parameters
    ----------
    seqids : list
        List with the sequence identifiers of the sequences
        to be translated.
    protein_file : str
        Path to a file to save protein sequences.
    sequences_file : str
        Path to the FASTA file that contains the DNA sequences.
    translation_table : int
        Translation table identifier.
    minimum_length : int
        The minimum sequence length value.

    Returns
    -------
    A list with following elements:
        invalid_alleles : list
            List with one sublist per invalid allele.
            Each sublist contains a sequence identifer
            and the exception message returned after
            attempting translation.
        total_seqs : int
            Total number of DNA sequences that were
            translated.
    """
    # define limit of records to keep in memory
    total_seqs = 0
    prot_lines = []
    line_limit = 10000
    invalid_alleles = []
    cds_index = fao.index_fasta(sequences_file)
    for i, seqid in enumerate(seqids):
        sequence = str(cds_index.get(seqid).seq)

        translation = translate_dna(sequence, translation_table, minimum_length)
        if isinstance(translation, list):
            prot_lines.append('>{0}'.format(seqid))
            prot_lines.append(str(translation[0][0]))
            total_seqs += 1
        # if returned value is a string, translation failed and
        # string contains exceptions
        elif isinstance(translation, str):
            invalid_alleles.append([seqid, translation])

        if len(prot_lines)//2 == line_limit or i+1 == len(seqids):
            prot_lines = im.join_list(prot_lines, '\n')
            fo.write_to_file(prot_lines, protein_file, 'a', '\n')
            prot_lines = []

    return [invalid_alleles, total_seqs]


def determine_distinct(sequences_file, unique_fasta, map_ids):
    """Identify duplicated sequences in a FASTA file.

    Parameters
    ----------
    sequences_file : str
        Path to a FASTA file.
    unique_fasta : str
        Path to a FASTA file that will be created to
        store distinct sequences.
    map_ids : dict
        Dictionary with mapping between genome string
        identifiers and genome integer identifiers.

    Returns
    -------
    pickle_out : str
        Pickled file that contains a dictionary with sequence
        hashes as keys and a list with pairs of protein
        identifiers and genome identifiers as values
    """
    out_seqs = []
    duplicates = {}
    exausted = False
    # limit of 10000 Fasta records in memory
    out_limit = 10000
    seq_generator = fao.sequence_generator(sequences_file)
    while exausted is False:
        record = next(seq_generator, None)
        if record is not None:
            # seq object has to be converted to string
            seqid = record.id
            sequence = str(record.seq.upper())

            # use digest() instead of hexdigest() to reduce memory usage?
            seq_hash = im.hash_sequence(sequence)

            # add unseen sequence to Fasta file with distinct sequences
            if seq_hash not in duplicates:
                recout = fao.fasta_str_record(ct.FASTA_RECORD_TEMPLATE, [seqid, sequence])
                out_seqs.append(recout)

            # add CDS hash as key
            # add genome integer identifier and protein identifier to values list
            # genome identifier and protein identifier can be used to fetch sequences
            genome_id, protid = seqid.split('-protein')
            genome_id = map_ids[genome_id]
            duplicates.setdefault(seq_hash, []).extend([int(protid), int(genome_id)])
        else:
            exausted = True

        # write Fasta records to file
        if len(out_seqs) == out_limit or exausted is True:
            if len(out_seqs) > 0:
                out_seqs = im.join_list(out_seqs, '\n')
                fo.write_to_file(out_seqs, unique_fasta, 'a', '\n')
                # reset list to avoid writing same records multiple times
                out_seqs = []

    # save dictionary with genome integer identifiers per distinct sequence
    # to pickle and only return file path to avoid keeping all dicts from
    # parallel processes in memory
    pickle_out = unique_fasta.replace('.fasta', '.duplicates')
    fo.pickle_dumper(duplicates, pickle_out)

    return pickle_out


def determine_small(sequences_file, minimum_length, variation=0):
    """Find protein sequences that are shorter than a specified length.

    Parameters
    ----------
    sequences_file : str
        Path to a FASTA file.
    minimum_length : int
        Sequences with a length value below this value
        are considered small.
    variation : float
        Accept sequences with length variation of up to
        minus (`minimum_length`*`variation`).

    Returns
    -------
    small_seqids : list
        List with the identifiers of small sequences.
    """
    variation = minimum_length - (minimum_length*variation)
    small_seqids = []
    seq_generator = fao.sequence_generator(sequences_file)
    for record in seq_generator:
        # seq object has to be converted to string
        sequence = str(record.seq)
        seqid = record.id

        if len(sequence) < variation:
            small_seqids.append(seqid)

    return small_seqids


def apply_bsr(blast_results, fasta_file, bsr, ids_dict):
    """Find similar sequences based on the BLAST Score Ratio.

    Parameters
    ----------
    blast_results : list
        List with the path to a file with BLAST
        results in tabular format.
    fasta_file : str
        Path to a FASTA file that contains the
        sequences that were aligned.
    bsr : float
        The BSR value to use as threshold
    ids_dict : dict
        Dictionary with the mapping between
        sequence identifiers used for BLAST and
        the original sequence identifiers.

    Returns
    -------
    excluded_alleles : list
        List with the identifiers of the sequences
        that were highly similar to other sequences.
    """
    self_scores = {r[0]: r[-1] for r in blast_results if r[0] == r[4]}
    # do not include self-scores lines, no need to evaluate those hits
    blast_results = [r for r in blast_results if r[0] != r[4]]

    lengths = {}
    for k in self_scores:
        record = fasta_file.get(ids_dict[k])
        sequence = str(record.seq)
        lengths[k] = len(sequence)

    excluded_alleles = []
    for res in blast_results:

        query = res[0]
        hit = res[4]
        score = res[-1]

        if query not in excluded_alleles:
            # try to apply BSR strategy
            try:
                self_blast_score = self_scores[query]

                query_length = lengths[query]
                hit_length = lengths[hit]
                blast_score_ratio = float(score) / float(self_blast_score)

                # BSR has to be greater than threshold, just as in the original function
                if blast_score_ratio >= bsr and hit not in excluded_alleles:

                    if hit_length > query_length and query not in excluded_alleles:
                        excluded_alleles.append(query)

                    elif hit_length <= query_length:
                        excluded_alleles.append(hit)
            # it might not work because there is no self score for
            # some sequences due to low complexity regions...
            except Exception:
                excluded_alleles.append(query)

    return excluded_alleles



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module enables the creation of a TSV file with annotation
terms for the loci in a schema.

The process queries UniProt's SPARQL endpoint to find exact
matches and retrieve the product name and page URL for those
matches. If users provide a taxon/taxa name/s, the process
will also search for reference proteomes for the specified
taxon/taxa and use BLASTp to align local sequences against
reference sequences to assign annotation terms based on the
BSR value computed for each alignment.

Code documentation
------------------
"""


import os
import sys

try:
    from utils import (
        constants as ct,
        blast_wrapper as bw,
        core_functions as cf,
        file_operations as fo,
        uniprot_requests as ur,
        fasta_operations as fao,
        parameters_validation as pv,
        iterables_manipulation as im,
        multiprocessing_operations as mo)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (
        constants as ct,
        blast_wrapper as bw,
        core_functions as cf,
        file_operations as fo,
        uniprot_requests as ur,
        fasta_operations as fao,
        parameters_validation as pv,
        iterables_manipulation as im,
        multiprocessing_operations as mo)


def extract_annotations(blastout_files, indexed_proteome, self_scores,
                        blast_score_ratio, proteome_matches):
    """Extract the annotation terms from high-scoring proteome matches.

    Parameters
    ----------
    blastout_files : list
        List with the paths to the TSV files withBLASTp results
        in tabular format. One per locus.
    indexed_proteome : Bio.File._IndexedSeqFileDict
        Fasta file index created with BioPython.
    self_scores : dict
        Dictionary with the identifiers of schema representatives
        as keys and the self-alignment raw socre as value.
    blast_score_ratio : float
        BLAST Score Ratio value. Hits with a BSR value
        >= than this value will be considered as high
        scoring hits that can be included in the final
        table according to the maximum number of matches
        to report.
    proteome_matches : int
        Maximum number of proteome matches to report.

    Returns
    -------
    proteome_results : dict
        Dictionary with loci identifiers as keys and a list
        with information about loci retrieved from the most
        similar records in UniProt's reference proteomes.
    """
    proteome_results = {}
    for file in blastout_files:
        locus_id = fo.file_basename(file).split('_short')[0]
        results = fo.read_tabular(file)
        if len(results) > 0:
            # compute BSR values
            for r in results:
                r.append(round(float(r[6])/float(self_scores[r[0]][1]), 2))

            # sort based on decreasing BSR
            sorted_results = sorted(results, key=lambda x: x[-1], reverse=True)
            # get results equal or above BSR
            high_bsr_results = [r for r in sorted_results
                                if r[-1] >= blast_score_ratio]

            for res in high_bsr_results[0:proteome_matches]:
                # get record and extract relevant info
                hit = indexed_proteome[res[4]]
                hit_dict = vars(hit)
                hit_terms = ur.extract_proteome_terms(hit_dict)
                proteome_results.setdefault(locus_id, []).append(hit_terms+[res[-1]])

    return proteome_results


def proteome_annotations(schema_directory, temp_directory, taxa,
                         blast_score_ratio, cpu_cores, proteome_matches,
                         blast_path):
    """Get annotations based on matches against UniProt's reference proteomes.

    Determine loci annotations based on alignment against UniProt's
    reference proteomes.

    Parameters
    ----------
    schema_directory : str
        Path to the schema's directory.
    temp_directory : str
        Path to the temporary directory where intermediate
        files will be written to.
    taxa : list
        List of taxa scientific names. The process will
        search for reference proteomes whose "Species Name"
        field contain any of the provided taxa names.
    blast_score_ratio : float
        BLAST Score Ratio value. Hits with a BSR value
        >= than this value will be considered as high
        scoring hits that can be included in the final
        table according to the maximum number of matches
        to report.
    cpu_cores : int
        Number of threads used to run BLASTp.
    proteome_matches : int
        Maximum number of proteome matches to report.
    blast_path : str
        Path to BLAST executables.

    Returns
    -------
    proteome_results : dict
        Dictionary with loci identifiers as keys and a list
        with information about loci retrieved from the most
        similar records in UniProt's reference proteomes.
    """
    # get paths to files with representative sequences
    short_directory = fo.join_paths(schema_directory, ['short'])
    reps_paths = [fo.join_paths(short_directory, [file])
                  for file in os.listdir(short_directory)
                  if file.endswith('.fasta') is True]

    print('Translating representative sequences...')
    # translate representatives for all loci
    translated_reps = fo.join_paths(temp_directory, ['translated_reps'])
    fo.create_directory(translated_reps)

    reps_protein_files = mo.parallelize_function(fao.translate_fasta,
                                                 reps_paths,
                                                 [translated_reps, 11],
                                                 cpu_cores, True)
    reps_protein_files = [r[1] for r in reps_protein_files]

    print('\nDownloading list of reference proteomes...', end='')
    remote_readme = fo.join_paths(ct.UNIPROT_PROTEOMES_FTP, ['README'])
    local_readme = fo.join_paths(temp_directory,
                                 ['reference_proteomes_readme.txt'])

    # get README file with list of reference proteomes
    res = fo.download_file(remote_readme, local_readme)
    print('done.')

    # get lines with proteomes info for species of interest
    readme_lines = fo.read_lines(local_readme, strip=False)

    selected_proteomes = im.contained_terms(readme_lines, taxa)
    selected_proteomes = [line.strip('\n') for line in selected_proteomes]
    selected_proteomes = [line.split('\t') for line in selected_proteomes]
    print('Found {0} reference proteomes for '
          '{1}.'.format(len(selected_proteomes), taxa))
    proteome_results = {}
    if len(selected_proteomes) > 0:
        # create directory to store proteomes
        proteomes_directory = fo.join_paths(temp_directory, ['proteomes'])
        fo.create_directory(proteomes_directory)

        proteomes_files = ur.get_proteomes(selected_proteomes,
                                           proteomes_directory)

        # uncompress files and concatenate into single FASTA
        uncompressed_proteomes = [fo.unzip_file(file) for file in proteomes_files]
        proteomes_concat = fo.join_paths(proteomes_directory,
                                         ['full_proteome.fasta'])
        proteomes_concat = fo.concatenate_files(uncompressed_proteomes,
                                                proteomes_concat)

        # get self-scores
        # concatenate protein files
        reps_concat = fo.concatenate_files(reps_protein_files,
                                           fo.join_paths(temp_directory,
                                                         ['reps_concat.fasta']))

        print('\nDetermining self-score of representatives...', end='')
        blastp_path = os.path.join(blast_path, ct.BLASTP_ALIAS)
        makeblastdb_path = os.path.join(blast_path, ct.MAKEBLASTDB_ALIAS)
        self_scores = cf.determine_self_scores(reps_concat, temp_directory,
            makeblastdb_path, blastp_path, 'prot', cpu_cores)
        print('done.')

        # create BLASTdb with proteome sequences
        proteome_blastdb = fo.join_paths(proteomes_directory,
                                         ['proteomes_db'])
        stderr = bw.make_blast_db('makeblastdb', proteomes_concat,
                                  proteome_blastdb, 'prot')

        # BLASTp to determine annotations
        blast_inputs = [['blastp', proteome_blastdb, file, file+'_blastout.tsv',
                         1, 1, None, None, proteome_matches, None, None, bw.run_blast]
                        for file in reps_protein_files]

        print('\nBLASTing representatives against proteomes...')
        blast_results = mo.map_async_parallelizer(blast_inputs,
                                                  mo.function_helper,
                                                  cpu_cores,
                                                  show_progress=True)

        blastout_files = [fo.join_paths(translated_reps, [file])
                          for file in os.listdir(translated_reps)
                          if 'blastout' in file]

        # index proteome file
        indexed_proteome = fao.index_fasta(proteomes_concat)

        # process results for each BLASTp
        proteome_results = extract_annotations(blastout_files,
                                               indexed_proteome,
                                               self_scores,
                                               blast_score_ratio,
                                               proteome_matches)

    return proteome_results


def sparql_annotations(loci_files, translation_table):
    """Retrieve annotations from UniProt's SPARQL endpoint.

    Parameters
    ----------
    loci_files : list
        List with the paths to the loci FASTA files.

    Returns
    -------
    annotations : list
        List with sublists. Each sublist contains
        the path to the FASTA file of a locus, the
        product name found for that locus and the
        URL to the page of the record that matched the
        locus.
    """
    # create inputs to multiprocessing
    uniprot_args = [[gene, translation_table, ur.get_annotation]
                    for gene in loci_files]

    # this works with all alleles in the loci to maximize
    # chance of finding info
    workers = ct.UNIPROT_SPARQL_THREADS
    annotations = mo.map_async_parallelizer(uniprot_args,
                                            mo.function_helper,
                                            workers,
                                            show_progress=True)

    return annotations


def create_annotations_table(annotations, output_directory, header,
                             schema_name):
    """Create output table with loci information.

    Parameters
    ----------
    annotations : dcit
        Dictionary with loci identifiers as keys and
        lists with information about loci as values (each
        list contains the information extracted from the
        "cds_info.tsv" table, if it was passed to the process,
        and the product and URL link for the match found
        through UniProt's SPARQL endpoint).
    output_directory : str
        Path to the output directory where the table
        will be written to.
    header : list
        File header (first line with column names).
    schema_name : str
        Name of the schema.
    loci_info : bool
        True if the user passed the "cds_info.tsv" table
        to the process, false otherwise.

    Returns
    -------
    output_table : str
        Path to the table with loci information.
    """
    annotation_lines = [header]
    for locus, data in annotations.items():
        locus_annotations = [locus]
        if isinstance(data[-1], list) is False:
            locus_annotations.extend(data)
        else:
            locus_annotations.extend(data[:-1])
            proteome_data = list(zip(*data[-1]))
            proteome_data = [';'.join(list(map(str, d))) for d in proteome_data]
            proteome_data = ['' if set(d) == {';'} else d for d in proteome_data]
            locus_annotations.extend(proteome_data)

        annotation_lines.append(locus_annotations)

    annotation_lines = ['\t'.join(line) for line in annotation_lines]
    output_basename = '{0}_annotations.tsv'.format(schema_name)
    output_file = fo.join_paths(output_directory, [output_basename])
    fo.write_lines(annotation_lines, output_file)

    return output_file


def main(input_files, output_directory, protein_table, blast_score_ratio,
         cpu_cores, taxa, proteome_matches, no_sparql, no_cleanup, blast_path):

    # create output directory
    fo.create_directory(output_directory)

    # create temp directory
    temp_directory = fo.join_paths(output_directory, ['temp'])
    fo.create_directory(temp_directory)

    # validate input files
    genes_list = fo.join_paths(temp_directory, [ct.LOCI_LIST])
    genes_list = pv.check_input_type(input_files, genes_list)
    loci_paths = fo.read_lines(genes_list)
    loci_basenames = [fo.file_basename(locus, False) for locus in loci_paths]

    schema_directory = os.path.dirname(loci_paths[0])
    schema_basename = fo.file_basename(schema_directory)
    print('Schema: {0}'.format(schema_directory))
    print('Number of loci: {0}'.format(len(loci_paths)))

    # find annotations based on reference proteomes for species
    proteome_results = {}
    if taxa is not None:
        proteome_results = proteome_annotations(schema_directory,
                                                temp_directory,
                                                taxa,
                                                blast_score_ratio,
                                                cpu_cores,
                                                proteome_matches,
                                                blast_path)
    else:
        print('No taxa names provided. Will not annotate based on '
              'UniProt\'s reference proteomes.')

    failed = {}
    sparql_results = {}
    if no_sparql is False:
        # check if SPARQL endpoint is up
        available = ur.website_availability(ct.UNIPROT_SPARQL)
        if available.code != 200:
            print('Cannot retrieve annotations from UniProt\'s SPARQL endpoint.')
            print(str(available))
        else:
            # search for annotations through the SPARQL endpoint
            print('\nQuerying UniProt\'s SPARQL endpoint...')
            config_file = fo.join_paths(input_files, ['.schema_config'])
            if os.path.isfile(config_file) is True:
                config = fo.pickle_loader(config_file)
                translation_table = config.get('translation_table', [11])[0]
            else:
                translation_table = 11

            # get annotations through UniProt SPARQL endpoint
            results = sparql_annotations(loci_paths, translation_table)

            sparql_results = {fo.file_basename(r[0], False): r[1:-1]
                              for r in results}
            found = sum([1 for k, v in sparql_results.items() if set(v) != {''}])
            print('\nFound annotations for {0}/{1} loci.'.format(found, len(loci_paths)))

            failed = {fo.file_basename(r[0], False): r[-1]
                      for r in results if len(r[-1]) > 0}
    else:
        print('\nProvided "--no-sparql" argument. Skipped step to '
              'search for annotations through UniProt\'s SPARQL '
              'endpoint.')

    if len(proteome_results) == 0 and len(sparql_results) == 0:
        exists = fo.delete_directory(output_directory)
        sys.exit('Could not retrieve annotations for any loci.')
    else:
        header = ['Locus']
        loci_info = {}
        if protein_table is not None:
            # read "cds_coordinates.tsv" file created by CreateSchema
            table_lines = fo.read_tabular(protein_table)
            header += table_lines[0]
            for l in table_lines[1:]:
                # create locus identifier based on genome identifier and
                # cds identifier in file
                locus_id = l[0].replace('_', '-')
                locus_id = locus_id + '-protein{0}'.format(l[-2])
                loci_info[locus_id] = l

        if no_sparql is False:
            header += ['Uniprot_Name', 'UniProt_URL']

        if taxa is not None:
            header.extend(['Proteome_ID', 'Proteome_Product',
                           'Proteome_Gene_Name', 'Proteome_Species',
                           'Proteome_BSR'])

        annotations = {}
        for locus in loci_basenames:
            annotations[locus] = []
            if protein_table is not None:
                annotations[locus] += loci_info.get(locus, ['']*6)

            if no_sparql is False:
                annotations[locus] += sparql_results.get(locus, ['']*2)

            if taxa is not None:
                annotations[locus].append(proteome_results.get(locus, [['']*5]))

        output_file = create_annotations_table(annotations, output_directory,
                                               header, schema_basename)

        print('\nThe table with new information can be found at:'
              '\n{0}'.format(output_file))

        # write file with information about cases that failed
        if len(failed) > 0:
            failed_lines = []
            for locus, messages in failed.items():
                distinct_messages = list(set([m.msg for m in messages]))
                locus_message = '{0}:\n{1}'.format(locus, '\n'.join(distinct_messages))
                failed_lines.append(locus_message)
            failed_outfile = fo.join_paths(output_directory, ['failed.txt'])
            failed_text = '\n'.join(failed_lines)
            fo.write_to_file(failed_text, failed_outfile, 'w', '\n')

        if no_cleanup is False:
            exists = fo.delete_directory(temp_directory)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------
This module enables the adaptation of external schemas so that the loci and
alleles present in those schemas can be used with chewBBACA. During the
process, alleles that do not correspond to a complete CDS or that cannot be
translated are discarded from the final schema. One or more alleles of each
gene/locus will be chosen as representatives and included in the 'short'
directory.

Expected input
--------------

The process expects the following variables whether through command line
execution or invocation of the :py:func:`main` function:

- ``-i``, ``input_files`` : Path to the folder containing the fasta files,
  one fasta file per gene/locus (alternatively, a file with a list of paths
  can be given).

    - e.g.: ``/home/user/chewie_schemas/schema_dir``

- ``-o``, ``output_directory`` : The directory where the output files will
  be saved (will create the directory if it does not exist).

    - e.g.: ``/home/user/adapted_schema``

- ``ptf``, ``ptf_path`` : Path to the Prodigal training file that will
  be associated with the adapted schema.

    - e.g.: ``/home/user/training_files/training_file``

- ``--cpu``, ``cpu_cores`` : The number of CPU cores to use (default=1).

    - e.g.: ``4``

- ``--bsr``, ``blast_score_ratio`` : The BLAST Score Ratio value that
  will be used to adapt the external schema (default=0.6).

    - e.g.: ``0.6``

- ``--l``, ``minimum_length`` : Minimum sequence length accepted.
  Sequences with a length value smaller than the value passed to this
  argument will be discarded (default=0).

    - e.g.: ``201``

- ``--t``, ``translation_table`` : Genetic code to use for CDS
  translation (default=11, for Bacteria and Archaea).

    - e.g.: ``11``

- ``--st``, ``size_threshold`` : CDS size variation threshold. At the
  default value of 0.2, alleles with size variation +-20 percent when
  compared to the representative will not be included in the final schema.

    - e.g.: ``0.2``

Code documentation
------------------
"""

import os
import shutil
import argparse
import itertools
import multiprocessing

try:
    from utils import (constants as ct,
                       blast_wrapper as bw,
                       file_operations as fo,
                       fasta_operations as fao,
                       sequence_manipulation as sm,
                       parameters_validation as pv,
                       iterables_manipulation as im,
                       multiprocessing_operations as mo)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 blast_wrapper as bw,
                                 file_operations as fo,
                                 fasta_operations as fao,
                                 sequence_manipulation as sm,
                                 parameters_validation as pv,
                                 iterables_manipulation as im,
                                 multiprocessing_operations as mo)


def bsr_categorizer(blast_results, representatives,
                    representatives_scores, min_bsr, max_bsr):
    """Identify BLAST hits below and above the BSR min and max thresholds.

    Parameters
    ----------
    blast_results : list of list
        A list with sublists, each sublist contains information
        about a BLAST hit.
    representatives : list
        List with sequence identifiers of representative
        sequences.
    representatives_scores : dict
        Dictionary with self BLAST raw score for every
        representative.
    min_bsr : float
        Minimum BSR value accepted to consider a sequence as
        a possible new representative.
    max_bsr : float
        Maximum BSR value accepted to consider a sequence as
        a possible new representative.

    Returns
    -------
    List with the following elements:
        high_bsr : list
            list with all sequence identifiers of subject
            sequences that had hits with a BSR higher than the
            maximum defined threshold.
        low_bsr : list
            list with all sequence identifiers of subject
            sequences that had hits with a BSR lower than the
            minimum defined threshold.
    """
    high_bsr = []
    hotspot_bsr = []
    low_bsr = []

    high_reps = {}
    hot_reps = {}
    low_reps = {}

    filtered_results = [res for res in blast_results
                        if res[0] != res[4] and res[4] not in representatives]
    bsr_values = [float(res[-1])/float(representatives_scores[res[0]])
                  for res in filtered_results]

    high_bsr = [res[4] for ind, res in enumerate(filtered_results)
                if bsr_values[ind] >= max_bsr]
    low_bsr = [res[4] for ind, res in enumerate(filtered_results)
               if bsr_values[ind] < min_bsr]
    hotspot_bsr = [res[4] for ind, res in enumerate(filtered_results)
                   if bsr_values[ind] >= min_bsr and bsr_values[ind] < max_bsr]

    for ind, res in enumerate(filtered_results):
        if bsr_values[ind] >= min_bsr:
            high_reps.setdefault(res[0], []).append(res[4])
        if bsr_values[ind] < min_bsr:
            low_reps.setdefault(res[0], []).append(res[4])
        if bsr_values[ind] >= min_bsr and bsr_values[ind] < max_bsr:
            hot_reps.setdefault(res[0], []).append(res[4])

    # determine representatives that only led to low BSR
    low_reps = list(set(low_reps) - set(high_reps))

    return [high_bsr, low_bsr, hotspot_bsr, high_reps, low_reps, hot_reps]


def select_candidate(candidates, proteins, seqids,
                     representatives, final_representatives):
    """Select a new representative sequence.

    Parameters
    ----------
    candidates : list
        List with the sequence identifiers of all candidates.
    proteins : dict
        A dictionary with protein identifiers as keys and
        protein sequences as values.
    seqids : list
        A list with the sequence identifiers that still have
        no representative (representatives identifiers are
        included because they have to be BLASTed in order to
        determine their self score).
    representatives : list
        The sequence identifiers of all representatives.

    Returns
    -------
    representatives : list
        The set of all representatives, including the new
        representative that was chosen by the function.
    """
    # with more than one sequence as candidate, select longest
    if len(candidates) > 1:

        # determine length of all candidates
        candidates_len = [(seqid, len(proteins[seqid]))
                          for seqid in candidates]

        # order representative candidates by length descending order
        candidates_len = sorted(candidates_len, key=lambda x: x[1],
                                reverse=True)

        # longest allele is the new representative
        representatives.append(candidates_len[0][0])
        final_representatives.append(candidates_len[0][0])

    # if there is only one candidate, keep that
    elif len(candidates) == 1:

        representatives.append(candidates[0])
        final_representatives.append(candidates[0])

    # if no hit qualifies and there are still sequences
    # without representative
    elif len(candidates) == 0 and \
            len(seqids) > len(representatives):

        # determine length of remaining sequences
        # (representatives not included)
        candidates_len = [(seqid, len(proteins[seqid]))
                          for seqid in seqids
                          if seqid not in representatives]

        # sort by descending length
        candidates_len = sorted(candidates_len, key=lambda x: x[1],
                                reverse=True)

        # longest of remaining sequences is new representative
        representatives.append(candidates_len[0][0])
        final_representatives.append(candidates_len[0][0])

    return [representatives, final_representatives]


def adapt_loci(genes, schema_path, schema_short_path, bsr, min_len,
               table_id, size_threshold, blastp_path, makeblastdb_path):
    """Adapts a set of genes/loci from an external schema.

    Adapts an external schema for usage with chewBBACA. Removes invalid
    alleles and selects representative alleles to include in the "short"
    directory.

    Parameters
    ----------
    genes_list : list
        A list with the following elements:

        - List with paths to the files to be processed.
        - Path to the schema directory.
        - Path to the "short" directory.
        - BLAST Score Ratio value.
        - Minimum sequence length value.
        - Genetic code.
        - Sequence size variation threshold.

    Returns
    -------
    invalid_alleles : list
        List with the identifiers of the alleles that were
        determined to be invalid.
    invalid_genes : list
        List with the identifiers of the genes that had no
        valid alleles.
    summary_stats : list of list
        List with one sublist per processed locus. Each
        sublist has four elements:

        - The identifier of the locus.
        - The number of alleles in the external file.
        - The number of alleles that were a valid CDS.
        - The number of representatives determined determined
          by the process.

    The function writes the schema files .
    """
    # divide input list into variables
    summary_stats = []
    invalid_genes = []
    invalid_alleles = []
    for gene in genes:

        representatives = []
        final_representatives = []

        # get gene basename and identifier
        gene_basename = os.path.basename(gene)
        gene_id = gene_basename.split('.fasta')[0]

        # create paths to gene files in new schema
        gene_file = fo.join_paths(schema_path, [gene_basename])

        gene_short_file = fo.join_paths(schema_short_path,
                                        [gene_basename.replace('.fasta', '_short.fasta')])

        # create path to temp working directory for current gene
        gene_temp_dir = fo.join_paths(schema_path,
                                      ['{0}{1}'.format(gene_id, '_temp')])

        # create temp directory for the current gene
        fo.create_directory(gene_temp_dir)

        # dictionaries mapping gene identifiers to DNA sequences
        # and Protein sequences
        gene_seqs, prot_seqs, gene_invalid, seqids_map, total_sequences = \
            sm.get_seqs_dicts(gene, gene_id, table_id, min_len, size_threshold)
        invalid_alleles.extend(gene_invalid)

        # if locus has no valid CDS sequences,
        # continue to next locus
        if len(prot_seqs) == 0:
            shutil.rmtree(gene_temp_dir)
            invalid_genes.append(gene_id)
            summary_stats.append([gene_id, str(total_sequences), '0', '0'])
            continue

        if len(gene_seqs) > 1:
            # identify DNA sequences that code for same protein
            equal_prots = sm.determine_duplicated_seqs(prot_seqs)

            # get only one identifier per protein
            ids_to_blast = [protids[0] for protein, protids in equal_prots.items()]

            # get longest sequence as first representative
            longest = sm.determine_longest(ids_to_blast, prot_seqs)
            representatives.append(longest)
            final_representatives.append(longest)

            # create FASTA file with distinct protein sequences
            protein_file = fo.join_paths(gene_temp_dir,
                                         ['{0}_protein.fasta'.format(gene_id)])
            protein_data = [[i, prot_seqs[i]] for i in ids_to_blast]
            protein_lines = fao.fasta_lines(ct.FASTA_RECORD_TEMPLATE, protein_data)
            fo.write_lines(protein_lines, protein_file)

            # create blastdb with all distinct proteins
            blastp_db = os.path.join(gene_temp_dir, gene_id)
            bw.make_blast_db(makeblastdb_path, protein_file, blastp_db, 'prot')

            # determine appropriate blastp task (proteins < 30aa need blastp-short)
            blastp_task = bw.determine_blast_task(equal_prots)

            # cycles to BLAST representatives against non-representatives until
            # all non-representatives have a representative
            while len(set(ids_to_blast) - set(representatives)) != 0:

                # create FASTA file with representative sequences
                rep_file = fo.join_paths(gene_temp_dir,
                                         ['{0}_rep_protein.fasta'.format(gene_id)])
                rep_protein_data = [[r, prot_seqs[r]] for r in representatives]
                rep_protein_lines = fao.fasta_lines(ct.FASTA_RECORD_TEMPLATE, rep_protein_data)
                fo.write_lines(rep_protein_lines, rep_file)

                # create file with seqids to BLAST against
                ids_str = im.join_list([str(i) for i in ids_to_blast], '\n')
                ids_file = fo.join_paths(gene_temp_dir,
                                         ['{0}_ids.txt'.format(gene_id)])
                fo.write_to_file(ids_str, ids_file, 'w', '')

                # BLAST representatives against non-represented
                blast_output = fo.join_paths(gene_temp_dir,
                                             ['{0}_blast_out.tsv'.format(gene_id)])
                # set max_target_seqs to huge number because BLAST only
                # returns 500 hits by default

                blast_stderr = bw.run_blast(blastp_path, blastp_db, rep_file,
                                            blast_output, 1, 1, ids_file,
                                            blastp_task, 100000, ignore=ct.IGNORE_RAISED)
                if len(blast_stderr) > 0:
                    raise ValueError(blast_stderr)

                # import BLAST results
                blast_results = fo.read_tabular(blast_output)

                # get self-score for representatives
                rep_self_scores = {res[0]: res[-1] for res in blast_results
                                   if res[0] == res[4]}

                # divide results into high, low and hot BSR values
                hitting_high, hitting_low, hotspots, high_reps, low_reps, hot_reps = \
                    bsr_categorizer(blast_results, representatives,
                                    rep_self_scores, bsr, bsr+0.1)

                excluded_reps = []

                # remove high BSR hits that have representative
                hitting_high = set(hitting_high)
                ids_to_blast = [i for i in ids_to_blast if i not in hitting_high]

                # remove representatives that led to high BSR with subjects that were removed
                prunned_high_reps = {k: [r for r in v if r in ids_to_blast] for k, v in high_reps.items()}
                reps_to_remove = [k for k, v in prunned_high_reps.items() if len(v) == 0]

                excluded_reps.extend(reps_to_remove)

                # determine smallest set of representatives that allow to get all cycle candidates
                excluded = []
                hotspot_reps = set(im.flatten_list(list(hot_reps.values())))
                for rep, hits in hot_reps.items():
                    common = hotspot_reps.intersection(set(hits))
                    if len(common) > 0:
                        hotspot_reps = hotspot_reps - common
                    else:
                        excluded.append(rep)

                excluded_reps.extend(excluded)

                # remove representatives that only led to low BSR
                excluded_reps.extend(low_reps)

                representatives = [rep for rep in representatives if rep not in excluded_reps]
                ids_to_blast = [i for i in ids_to_blast if i not in excluded_reps]

                # determine next representative from candidates
                rep_candidates = list(set(hotspots) - hitting_high)
                # sort to guarantee reproducible results with same datasets
                rep_candidates = sorted(rep_candidates, key=lambda x: int(x))
                representatives, final_representatives = select_candidate(rep_candidates,
                                                                          prot_seqs,
                                                                          ids_to_blast,
                                                                          representatives,
                                                                          final_representatives)

                # remove files created for current gene iteration
                os.remove(rep_file)
                os.remove(blast_output)
                os.remove(ids_file)

        else:
            final_representatives = list(prot_seqs.keys())

        # write schema file with all alleles
        gene_data = [[k, v] for k, v in gene_seqs.items()]
        gene_lines = fao.fasta_lines(ct.FASTA_RECORD_TEMPLATE, gene_data)
        fo.write_lines(gene_lines, gene_file)

        # get total number of valid sequences
        valid_sequences = len(gene_lines)

        # write schema file with representatives
        final_representatives = [seqids_map[rep] for rep in final_representatives]
        gene_rep_data = [[r, gene_seqs[r]] for r in final_representatives]
        gene_rep_lines = fao.fasta_lines(ct.FASTA_RECORD_TEMPLATE, gene_rep_data)
        fo.write_lines(gene_rep_lines, gene_short_file)

        # get number of representatives
        representatives_number = len(gene_rep_lines)

        summary_stats.append([gene_id,
                              str(total_sequences),
                              str(valid_sequences),
                              str(representatives_number)])

        shutil.rmtree(gene_temp_dir)

    return [invalid_alleles, invalid_genes, summary_stats]


def main(input_files, output_directory, cpu_cores, blast_score_ratio,
         minimum_length, translation_table, ptf_path, size_threshold,
         blast_path):

    print('Adapting schema in the following '
          'directory:\n{0}'.format(os.path.abspath(input_files)))
    print('Prodigal training file:\n{0}'.format(ptf_path))
    print('Number of cores: {0}'.format(cpu_cores))
    print('BLAST Score Ratio: {0}'.format(blast_score_ratio))
    print('Translation table: {0}'.format(translation_table))
    print('Minimum accepted sequence length: {0}'.format(minimum_length))
    print('Size threshold: {0}'.format(size_threshold))

    # define output paths
    schema_path = os.path.abspath(output_directory)
    schema_short_path = fo.join_paths(schema_path, ['short'])

    # create output directories
    # check if they exist first
    fo.create_directory(schema_path)
    fo.create_directory(schema_short_path)

    # list schema gene files
    genes_file = pv.check_input_type(input_files,
                                     fo.join_paths(output_directory, [ct.LOCI_LIST]))

    # import list of schema files
    with open(genes_file, 'r') as gf:
        genes_list = [line.rstrip('\n') for line in gf]
    os.remove(genes_file)

    print('Number of genes to adapt: {0}\n'.format(len(genes_list)))

    print('Determining the total number of alleles and '
          'allele mean length per gene...\n'.format())

    # count number of sequences and mean length per gene
    genes_info = []
    genes_pools = multiprocessing.Pool(processes=cpu_cores)
    gp = genes_pools.map_async(fao.fasta_stats, genes_list,
                               callback=genes_info.extend)
    gp.wait()

    # split files according to number of sequences and sequence mean length
    # in each file to pass even groups of sequences to all cores
    # divide into 100 input sets to get 1% progress resolution
    even_genes_groups = mo.distribute_loci(genes_info, 100, 'seqcount')
    # with few inputs, some sublists might be empty
    even_genes_groups = [i for i in even_genes_groups if len(i) > 0]
    # add common arguments
    blastp_path = os.path.join(blast_path, ct.BLASTP_ALIAS)
    makeblastdb_path = os.path.join(blast_path, ct.MAKEBLASTDB_ALIAS)
    even_genes_groups = [[i, schema_path, schema_short_path,
                          blast_score_ratio, minimum_length,
                          translation_table, size_threshold,
                          blastp_path, makeblastdb_path,
                          adapt_loci] for i in even_genes_groups]

    print('Adapting {0} genes...\n'.format(len(genes_list)))

    invalid_data = mo.map_async_parallelizer(even_genes_groups,
                                             mo.function_helper,
                                             cpu_cores,
                                             show_progress=True)

    # define paths and write files with list of invalid
    # alleles and invalid genes
    output_schema_basename = os.path.basename(output_directory.rstrip('/'))
    schema_parent_directory = os.path.dirname(schema_path)

    # write file with alleles that were determined to be invalid
    invalid_alleles = [sub[0] for sub in invalid_data]
    invalid_alleles = list(itertools.chain.from_iterable(invalid_alleles))
    invalid_alleles_file = os.path.join(schema_parent_directory,
                                        '{0}_{1}'.format(output_schema_basename, 'invalid_alleles.txt'))

    with open(invalid_alleles_file, 'w') as inv:
        lines = ['{0}: {1}\n'.format(allele[0], allele[1]) for allele in invalid_alleles]
        inv.writelines(lines)

    # write file with identifiers of genes that had no valid alleles
    invalid_genes = [sub[1] for sub in invalid_data]
    invalid_genes = list(itertools.chain.from_iterable(invalid_genes))
    invalid_genes_file = os.path.join(schema_parent_directory,
                                      '{0}_{1}'.format(output_schema_basename, 'invalid_genes.txt'))

    with open(invalid_genes_file, 'w') as inv:
        invalid_geqids = '\n'.join(invalid_genes)
        inv.write(invalid_geqids)

    stats_lines = [sub[2] for sub in invalid_data]
    stats_lines = list(itertools.chain.from_iterable(stats_lines))
    stats_lines = ['\t'.join(line) for line in stats_lines]
    stats_genes_file = '{0}/{1}_{2}'.format(schema_parent_directory,
                                            output_schema_basename,
                                            'summary_stats.tsv')

    with open(stats_genes_file, 'w') as stats:
        summary_stats_text = '\n'.join(stats_lines)
        stats.write('Gene\tTotal_alleles\tValid_alleles\tNumber_representatives\n')
        stats.write(summary_stats_text)

    print('\n\nNumber of invalid genes: {0}'.format(len(invalid_genes)))
    print('Number of invalid alleles: {0}'.format(len(invalid_alleles)))

    print('\nSuccessfully adapted {0}/{1} genes present in the '
          'input schema.'.format(len(genes_list)-len(invalid_genes),
                                 len(genes_list)))


def parse_arguments():

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-i', type=str, required=True, dest='input_files',
                        help='Path to the folder containing the fasta files, '
                             'one fasta file per gene/locus (alternatively, '
                             'a file with a list of paths can be given).')

    parser.add_argument('-o', type=str, required=True, dest='output_directory',
                        help='The directory where the output files will be '
                             'saved (will create the directory if it does not '
                             'exist).')

    parser.add_argument('--ptf', type=str, required=False,
                        default=False, dest='ptf_path',
                        help='Path to the Prodigal training file that '
                             'will be associated with the adapted schema.')

    parser.add_argument('--cpu', type=int, required=False, default=1,
                        dest='cpu_cores',
                        help='The number of CPU cores to use (default=1).')

    parser.add_argument('--bsr', type=float, required=False, default=0.6,
                        dest='blast_score_ratio',
                        help='The BLAST Score Ratio value that will be '
                             'used to adapt the external schema (default=0.6).')

    parser.add_argument('--l', type=int, required=False, default=0,
                        dest='minimum_length',
                        help='Minimum sequence length accepted. Sequences with'
                        ' a length value smaller than the value passed to this'
                        ' argument will be discarded (default=0).')

    parser.add_argument('--t', type=int, required=False, default=11,
                        dest='translation_table',
                        help='Genetic code to use for CDS translation.'
                        ' (default=11, for Bacteria and Archaea)')

    parser.add_argument('--st', type=float, required=True,
                        default=0.2, dest='size_threshold',
                        help='CDS size variation threshold. At the default '
                             'value of 0.2, alleles with size variation '
                             '+-20 percent when compared to the representative '
                             'will not be included in the final schema.')

    parser.add_argument('--b', type=pv.check_blast, required=False,
                        default='', dest='blast_path',
                        help='Path to the BLAST executables.')

    args = parser.parse_args()

    return args


if __name__ == '__main__':

    args = parse_arguments()

    main(**vars(args))
